package modules.binomial;

import java.awt.*;
import java.awt.geom.*;

import vcp.mathnotation.MathExpression;

public class DecisionTree {
  static final float NODE_SIZE = 8;
  static final float EDGE_LENGTH = 150;
  static final BasicStroke EDGE_NORMAL = new BasicStroke(2);
  static final BasicStroke EDGE_HIGHLIGHT = new BasicStroke(4);

  float spacing = 20;

  DecisionNode root;

  public DecisionTree(float m) {
    root = new DecisionNode(null);
    
    spacing = m;
  }

  public static class Point {
    float x, y;

    public Point(float xa, float ya) {
      x = xa;
      y = ya;
    }
  }

  public static class DecisionNode {
    DecisionNode parent;
    DecisionNode[] children;
    
    Color color;
    
    Color[] edgeColor;
    MathExpression[] edgeLabel;

    boolean highlight;

    Point l;

    public DecisionNode(DecisionNode p) {
      this(p,null);
    }

    public DecisionNode(DecisionNode p, DecisionNode[] c) {
      parent = p;
      children = c;

      color = Color.black;

      if(children != null) {
        edgeColor = new Color[children.length];
        edgeLabel = new MathExpression[children.length];
        //highlight = new boolean[children.length];
        for(int i = 0; i < edgeColor.length; i++)
          edgeColor[i] = Color.black;
      } else {
        edgeColor = null;
        edgeLabel = null;
        //highlight = null;
      }

    }

    public void highlightPath(boolean highlight) {
      DecisionNode current = this;

      //int l = 0;
      while(current != null) {
        //l++;
        current.highlight = highlight;
        current = current.parent;
      }
      //System.out.println("path length = " + l);
      
    }

    private void fixEdges() {
      if(children != null) {
        if(edgeColor == null) {
          edgeColor = new Color[children.length];
          for(int i = 0; i < edgeColor.length; i++)
            edgeColor[i] = Color.black;
        } else if(edgeColor.length != children.length) {
          Color[] temp = edgeColor;
          edgeColor = new Color[children.length];
          System.arraycopy(temp,0,edgeColor,0,temp.length);
          for(int i = temp.length; i < edgeColor.length; i++)
           edgeColor[i] = Color.black;            
        }
        
        if(edgeLabel == null) {
          edgeLabel = new MathExpression[children.length];
        } else if(edgeLabel.length != children.length) {
          MathExpression[] temp = edgeLabel;
          edgeLabel = new MathExpression[children.length];
          System.arraycopy(temp,0,edgeLabel,0,temp.length);
        }
        /*
        if(highlight == null) {
          highlight = new boolean[children.length];
        } else if(highlight.length != children.length) {
          boolean[] temp = highlight;
          highlight = new boolean[children.length];
          System.arraycopy(temp,0,highlight,0,temp.length);
        }
        */
      } else {
        edgeColor = null;
        edgeLabel = null;
        //highlight = null;
      }

    }

    public float childrenHeight(float spacing) {
      if(children[0].children == null) {
        return (children.length - 1) * spacing;
      }

      else {
        return (children[0].childrenHeight(spacing) * children.length) + 
          ((children.length - 1) * spacing);
      }
    }

    public float branchSpan(float spacing) {
      if(children[0].children == null) {
        return (children.length - 1) * spacing;
      }

      else {      
        return (children[0].childrenHeight(spacing) + spacing) * 
          (children.length - 1);
      }
    }

    public Arc2D.Float getArc() {
      return new Arc2D.Float
        (l.x-NODE_SIZE/2f,l.y-NODE_SIZE/2f,
         NODE_SIZE,NODE_SIZE,0,360,Arc2D.CHORD);
    }

    public void setEdge(int i, Color c) {
      fixEdges();

      edgeColor[i] = c;
    }

    public void setEdgeLabel(int i, MathExpression e) {
      fixEdges();

      edgeLabel[i] = e;
    }

    public void drawChildren(Graphics2D g, float spacing) {
      fixEdges();

      float s, h;

      //h = (children.length > 1) ? childrenHeight(spacing) : 0;
      //s = (children.length > 1) ? h/(children.length) : 0;
      h = branchSpan(spacing);
      s = h/(children.length-1);
      
      float y = l.y - (h / 2f);
      float x = l.x+EDGE_LENGTH;
      for(int i = 0; i < children.length; i++) {
        children[i].l = new Point(x,y);

        if(children[i].highlight) {
          //System.out.println("highlight! " + i);
          g.setColor(edgeColor[i].brighter().brighter());
          g.setStroke(EDGE_HIGHLIGHT);
        } else {
          g.setColor(edgeColor[i].darker());
          g.setStroke(EDGE_NORMAL);
        }
        
        g.draw(new Line2D.Float(l.x,l.y,x,y));

        if(edgeLabel[i] != null) {
          //edgeLabel[i].setGraphics(g);
          float lx;
          float ly;
          //if((y - l.y) > 0)
          //lx = l.x + (x - l.x)/2f + (float)edgeLabel[i].getWidth(g);
            //else
          lx = l.x + (x - l.x)/2f - (float)edgeLabel[i].getWidth(g);
          if((l.y - y) < 0)   
            ly = l.y + (y - l.y)/2f + (float)edgeLabel[i].getHeight(g);
          else 
            ly = l.y + (y - l.y)/2f;
          
          edgeLabel[i].draw(g,lx,ly);          
        }

        if(highlight)
          g.setColor(color.brighter().brighter());
        else
          g.setColor(color.darker());
        g.fill(getArc());

        if(children[i].children != null) {
          if(children[i].highlight)
            g.setColor(children[i].color.brighter().brighter());
          else
            g.setColor(children[i].color.darker());

          g.fill(children[i].getArc());
        }
        
        y += s;
      }
      
      if(children[0].children != null) {
        for(int i = 0; i < children.length; i++) {
          children[i].drawChildren(g, spacing);
        }
      }

    }

    public int getLeavesHelper(DecisionNode[] l, int li) {
      if(children == null) {
        l[li] = this;
        return li + 1;
      }
      
      else {
        for(int i = 0; i < children.length; i++) {
          li = children[i].getLeavesHelper(l,li);
        }
        return li;
      }
    }
  }
    
  public void draw(Graphics g, float x, float y) {
    Graphics2D g2 = (Graphics2D) g.create();

    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                        RenderingHints.VALUE_ANTIALIAS_ON);


    root.l = new Point(x,y);
    root.drawChildren(g2, spacing);
    if(root.highlight)
      g2.setColor(root.color.brighter().brighter());
    else
      g2.setColor(root.color.darker());

    g2.fill(root.getArc());
    

    g2.dispose();
  }

  public DecisionNode[] getLeaves(int n) {
    DecisionNode[] l = new DecisionNode[n];
    
    root.getLeavesHelper(l,0);

    return l;    
  }
  

  
}