package vcp;

import javax.swing.JFrame;
import javax.swing.JApplet;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JSeparator;
import javax.swing.KeyStroke;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import vcp.platform.*;

/**
 * This is the main class of the VCP application. It creates the main
 * JFrame and main content panel.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class VCP extends JApplet implements ActionListener {
  /** the main frame */
  private JFrame frame;

  /** true only if the class is running as an applet */
  private boolean applet = false;

  /** menu choice to load a module */
  private JMenuItem loadModule;

  /** menu choice to quit */
  private JMenuItem quit;

  /** menu choice to display about info */
  private JMenuItem about;

  private JMenuItem aboutVCP;

  private JMenuItem aboutComb;

  /** menu choice to display help information */
  private JMenuItem help;

  /** can we have this? */
  private JMenuItem pdfPrefs;

  /** the file menu */
  private JMenu file;

  /** main panel to display content */
  private JPanel mainPanel;

  /** a button used in the applet to show the main frame to the user
   * (in case the user accidentally closes the window or something) */
  private JButton showFrame;

  /**
   * Creates a new <code>VCP</code> instance with no default frame
   */
  public VCP() {
    this(null, true);
  }

  /**
   * Creates a new <code>VCP</code> instance with a default frame.
   *
   * @param f the default frame
   */
  public VCP(JFrame f, boolean isApplet) {
    applet = isApplet;

    frame = f;

    createMenu();

    mainPanel = new JPanel();
    mainPanel.setLayout(new java.awt.GridLayout(1,1));

    mainPanel.add(new ModulePanel());
    //mainPanel.add(new SplashGUI());

    GUIActions.initialize(this);
    ModuleChooser.initialize();
  }

  /**
   * If frame is not null, create and add menus.
   */
  private void createMenu() {
    if(frame != null) {
      JMenuBar menuBar;

      //Create the menu bar.
      menuBar = new JMenuBar();

      //Build the first menu.
      file = new JMenu("File");
      file.setMnemonic(KeyEvent.VK_F);
      menuBar.add(file);

      loadModule = new JMenuItem("Load Module...",
                                 KeyEvent.VK_L);
      loadModule.setAccelerator
        (KeyStroke.getKeyStroke
         (KeyEvent.VK_L, ActionEvent.CTRL_MASK));
      loadModule.addActionListener(this);
      file.add(loadModule);
      
      pdfPrefs = new JMenuItem("PDF Preferences...",
                               KeyEvent.VK_P);
      pdfPrefs.addActionListener(this);

      file.add(pdfPrefs);

      file.add(new JSeparator());

      quit = new JMenuItem("Quit",
                           KeyEvent.VK_Q);
      quit.addActionListener(this);
      file.add(quit);

      JMenu helpM = new JMenu("Help");
      helpM.setMnemonic(KeyEvent.VK_H);
      menuBar.add(helpM);

      help = new JMenuItem("Help", KeyEvent.VK_H);
      help.setAccelerator
        (KeyStroke.getKeyStroke
         (KeyEvent.VK_F1,0));
      help.addActionListener(this);
      helpM.add(help);

      helpM.add(new JSeparator());

      about = new JMenuItem("About the Creators ...", KeyEvent.VK_A);
      about.addActionListener(this);
      helpM.add(about);

      helpM.add(new JSeparator());
      
      aboutVCP = new JMenuItem("About VCP (PDF)", KeyEvent.VK_V);
      aboutVCP.addActionListener(this);
      helpM.add(aboutVCP);

      aboutComb = new JMenuItem("About Combinatorial Proofs (PDF)", KeyEvent.VK_C);
      aboutComb.addActionListener(this);
      helpM.add(aboutComb);
      
      frame.setJMenuBar(menuBar);
    }    
  }

  /**
   * Returns true if class is running as an applet
   *
   * @return true if class is running as an applet
   */
  public boolean isApplet() {
    return applet;
  }

  /**
   * Returns the frame
   *
   * @return the frame
   */
  public JFrame getFrame() {
    return frame;
  }

  /**
   * Set the content panel of the frame
   *
   * @param p the JPanel to be the content
   */
  public void setMainPanel(JPanel p) {
    mainPanel.removeAll();
    mainPanel.add(p);
    //mainPanel.setPreferredSize(new java.awt.Dimension(200,200));
    mainPanel.validate();
    frame.pack();
    frame.validate();
    validate();
  }

  /**
   * If the frame is not null, pack it
   */
  public void pack() {
    if(frame != null) {
      frame.pack();
    }
  }

  /**
   * If the frame is not null, set the title
   *
   * @param title the title of the frame
   */
  public void setTitle(String title) {
    if(frame != null) {
      frame.setTitle(title);
    }
  }

  /**
   * Handles all the menu actions.
   *
   * @param e an <code>ActionEvent</code> value
   */
  public void actionPerformed(ActionEvent e) {
    Object component = e.getSource();

    if(component == quit) {
      if(applet)
        frame.dispose();
      else
        System.exit(0);
    }

    else if(component == loadModule) {
      GUIActions.loadModule();
    }

    else if(component == help) {
      (new Help()).setVisible(true);
    }

    else if(component == about) {
      (new About()).setVisible(true);
    }

    else if(component == showFrame) {
      if(frame != null) {
        frame.setVisible(true);
        frame.setExtendedState(JFrame.NORMAL);
      }
    }

    else if(component == aboutVCP) {
      GUIActions.displayPDF
        ("modules/definitions/ProjectIntro.pdf");
    }

    else if(component == aboutComb) {
      GUIActions.displayPDF
	("modules/definitions/CombProofIntro.pdf");      
    }

    else if(component == pdfPrefs) {
      (new PDFPrefs()).setVisible(true);
    }
  }

  /**
   * If we are opn a Sun JVM, this method will be called if an
   * uncaught throwable is thrown.
   *
   * @param t a <code>Throwable</code> value
   */
  public void handle(Throwable t) {
    vcp.utilities.ExceptionDialog.showExceptionDialog
      ("Error",
       "An error ocurred.",t);
  }

  /**
   * Set the frame and create the menus
   *
   * @param f a <code>JFrame</code> value
   */
  private void setFrame(JFrame f) {
    frame = f;
    createMenu();
  }

  /**
   * Initializes applet
   */
  public void init() {
    applet = true;
    
    showFrame = new JButton("Show VCP Main Window");
    showFrame.addActionListener(this);

    getContentPane().add(showFrame);
    
    setFrame(createGUIFrame(this));
    
    file.remove(pdfPrefs);
    
    pack();
    frame.validate();
  }

  /**
   * Create a GUI frame
   *
   * @param vcp The VCP object this frame will belong to. If this is
   * null, a new VCP object will be created.
   * @return the JFrame created
   */
  private static JFrame createGUIFrame(VCP vcp) {
    JFrame f = new JFrame("VCP");
    if(vcp == null) {
      f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      vcp = new VCP(f,false);
    } else {
      f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    }

    f.getContentPane().add(vcp.mainPanel);
    f.pack();
    f.setVisible(true);

    return f;
  }

  /**
   * Run the class as an application.
   *
   * @param args a <code>String[]</code> value
   */
  public static void main(String[] args) {
    System.setProperty("sun.awt.exception.handler","vcp.VCP");

    javax.swing.SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          createGUIFrame(null);
        }
      });    
  }
}