package vcp.animation;

//It'd be nice to used java's Semaphores if the user has 1.5
//import java.util.concurrent.Semaphore;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.GridLayout;

/**
 * This is a class which can be used for "step" animation. That is,
 * animation with a fixed amount of time between frames.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class Animation implements Runnable {
  /** the default delay time (millis) */
  private static final int DEFAULT_DELAY = 300;

  /** the maximum delay speed (millis) */
  private int maxDelay = 800;

  /** the minimum delay speed (millis) */
  private int minDelay = 50;

  /** the delay step */
  private int delayStep;

  /** the thread controlling the current animation part */
  protected Thread aniThread;

  /** the current delay time */
  protected int delay = DEFAULT_DELAY;

  /** the control panel */
  protected ControlPanel cPanel;
  
  /** true if the animation is paused */
  private boolean paused;

  /** semaphore controlling pausing */
  private Semaphore sPause;

  /** semaphore controlling stepping */
  private Semaphore sStep;

  /** the array of animation parts */
  protected AnimationPart[] partArray;

  /** the next part of the animation to execute */
  protected int part;

  /**
   * Creates a new <code>Animation</code> instance with no
   * ControlPanel and default delay timings.
   * 
   * <p>Default values
   * <br><code>delay = 300</code>
   * <br><code>minDelay = 50</code>
   * <br><code>maxDelay = 800</code></p>
   */
  protected Animation() {
    aniThread = null;
    paused = false;
    sPause = new Semaphore(1);
    sStep = new Semaphore(1);
  }

  /**
   * Creates a new <code>Animation</code> instance with default delay
   * timings.
   * 
   * <p>Default values
   * <br><code>delay = 300</code>
   * <br><code>minDelay = 50</code>
   * <br><code>maxDelay = 800</code></p>
   *
   * @param c the <code>ControlPanel</code> for this Animation
   */
  public Animation(ControlPanel c) {
    this(c, 800, 50, DEFAULT_DELAY);
  }

  /**
   * Creates a new <code>Animation</code> instance with default delay
   * timing bounds.
   *
   * <p>Default values
   * <br><code>minDelay = 50</code>
   * <br><code>maxDelay = 800</code></p>
   * 
   * @param c the <code>ControlPanel</code> for this Animation
   * @param d the default delay time (millis)
   */
  public Animation(ControlPanel c, int d) {
    this(c, 800, 50, d);
  }

  /**
   * Creates a new <code>Animation</code> instance.
   *
   * @param c the <code>ControlPanel</code> for this Animation
   * @param mxd the maximum delay time (millis)
   * @param mnd the minimum delay time (millis)
   * @param d the default delay time (millis)
   */
  public Animation(ControlPanel c, int mxd, int mnd, int d) {
    aniThread = null;
    paused = false;
    sPause = new Semaphore(1);
    sStep = new Semaphore(1);

    delay = d;
    maxDelay = mxd;
    minDelay = mnd;
    
    if(delay > maxDelay)
      delay = maxDelay;
    else if(delay < minDelay)
      delay = minDelay;

    delayStep = (maxDelay - minDelay)/20;

    cPanel = c;
    cPanel.setAnimation(this);
    //cPanel.addActionListener(this);

    //partArray = createPartArray();

    part = 0;
  } 

  /**
   * Cause the animation to sleep for n milliseconds. This method is
   * used to control the speed of the animation. If the pause method
   * is called, it is at calls to sleep where the animation will
   * pause. The step method will step through the animation to the
   * next call to sleep.
   *
   * @param n the number of milliseconds to sleep
   * @see #sleep()
   */
  protected void sleep(int n) {
    try {
      Thread.sleep(n);
    } catch (InterruptedException e) {}
    
    sStep.acquireUninterruptibly();
    sPause.acquireUninterruptibly();
    sPause.release();
    sStep.release();
    synchronized(this) {}
  }

  /**
   * Cause the animation to sleep for the default number of
   * milliseconds.
   *
   * @see #sleep(int)
   */
  protected void sleep() {
    sleep(delay);
  }

  /**
   * Pauses the animation
   *
   * @see #start()
   * @see #step()
   */
  public void pause() {
    if(sPause.availablePermits() > 0) {
      sPause.acquireUninterruptibly();
      if(isInProgress()) aniThread.interrupt();
    }
  }

  /**
   * Starts the animation if it is paused
   *
   * @see #pause()
   * @see #step()
   */
  public void start() {
    sPause.release();
  }

  /**
   * If the animation is paused, step through the animation to the
   * next call to sleep.
   *
   * @see #pause()
   * @see #start()
   */
  public void step() {
    if(isPaused()) {
      sPause.release();
      synchronized(this) {
        sStep.acquireUninterruptibly();
        sPause.acquireUninterruptibly();
        aniThread.interrupt();
      }
      sStep.release();
    }
  }

  /**
   * Returns true if animation is currently in progress. This method
   * will return true whenever an {@link AnimationPart} is in progress
   * (even if the animation is paused).
   *
   * @return true if animation is currently in progress.
   * @see #isPaused()
   */
  public boolean isInProgress() {
    return aniThread != null;
  }

  /**
   * Returns true if animation is currently paused. There is no
   * guarantee on the return value of this method if no 
   * {@link AnimationPart} is in progress.
   *
   * @return true if animation is currently paused
   * @see #isInProgress()
   */
  public boolean isPaused() {
    return sPause.availablePermits() == 0;
  }

  /**
   * Slow the animation down one step.
   */
  public void slower() {
    delay += delayStep;
    
    if(delay > maxDelay)
      delay = maxDelay;

    //System.out.println(delay);
  }

  /**
   * Speed the animation up by one step
   */
  public void faster() {
    delay -= delayStep;

    if(delay < minDelay)
      delay = minDelay;

    //System.out.println(delay);
  }

  /**
   * Get the speed. Returns a relative value 0-100.
   *
   * @return the speed
   * @see #setSpeed(double)
   */
  public double getSpeed() {
    //System.out.println(delay);
    
    double value = (double)
      (
       100 - 
       100*(((double)(delay-minDelay)/
	     (double)(maxDelay-minDelay)))
       );
    
    //0% speed doesn't make much sense
    //if(percent == 0) percent = 1;

    return value;
  }

  /**
   * Set the speed of the animation.
   *
   * @param s a <code>double</code> value 0 - 100
   * @see #getSpeed()
   */
  public void setSpeed(double s) {
    delay = (int)(((100.0-s)/100.0)*(maxDelay-minDelay) + minDelay);
  }

  /**
   * Start the next part of the animation.
   *
   * @throws IllegalStateException if the AnimationPart array has not
   * yet been {@link #setPartArray(AnimationPart[]) set}.
   */
  public void startNextPart() {
    if(partArray == null)
      throw new IllegalStateException("part array has not been set");
    (new Thread(this)).start();
  }

  /** 
   * Executes the part of the animation corresponding to the part
   * field. This method is called each time the "next part" button is
   * pressed in the ControlPanel associated with this Animation.
   */
  public void run() {
    aniThread = Thread.currentThread();
    cPanel.partStarted();
    
    if(part < partArray.length)
      partArray[part++].animate();

    if(part < partArray.length)
      cPanel.partFinished();
    else
      cPanel.animationComplete();

    aniThread = null;
  }

  /**
   * Set the array of AnimationParts that make up this animation.
   *
   * @param pa the array of AnimationParts
   */
  public void setPartArray(AnimationPart[] pa) {
    partArray = pa;
  }

}