package vcp.animation;

import javax.swing.JComponent;

import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.RenderingHints;
import java.awt.Color;

import java.awt.image.BufferedImage;

/**
 * This is a canvas designed to be used for double-buffered
 * animations.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class AnimationCanvas extends JComponent {
  /** the buffer image */
  protected BufferedImage bimg;

  /** canvas width */
  protected int w = 600;

  /** canvas height */
  protected int h = 400;

  /**
   * Creates a new <code>AnimationCanvas</code> instance with the
   * default size 600x400 and default background color white.
   */
  public AnimationCanvas() {
    this(600,400,Color.white);
  }

  /**
   * Creates a new <code>AnimationCanvas</code> instance with the
   * default size 600x400.
   *
   * @param bg the background color
   */
  public AnimationCanvas(Color bg) {
    this(600,400,bg);
  }

  /**
   * Creates a new <code>AnimationCanvas</code> instance.
   *
   * @param width width in pixels
   * @param height height in pixels
   * @param bg the background color
   */
  public AnimationCanvas(int width, int height, Color bg) {
    super();

    w = width;
    h = height;

    setOpaque(true);

    Dimension d = new Dimension(w,h);
    setPreferredSize(d);
    setSize(d);
    setMinimumSize(d);
    setMaximumSize(d);
    setBackground(bg);
  }

  /**
   * Paints the contents of the buffer image to the canvas. Call this
   * method instead of repaint() to paint the canvas during
   * animations.
   */
  public void paintCanvas() {
    Graphics g;
    g = this.getGraphics();
    if((g != null) && (bimg != null)) {
      g.drawImage(bimg,0,0,null);
      g.dispose();
    }
  }

  public void paintComponent(Graphics g) {
    if(bimg != null)
      g.drawImage(bimg,0,0,null);
  }

  /**
   * Returns the graphics context for the buffer image. The image is
   * cleared first.
   *
   * @return the graphics context for the buffer image
   * @see #getImageGraphics(boolean)
   */
  public Graphics2D getImageGraphics() {
    return getImageGraphics(true);
  }

  /**
   * Returns the graphics context for the buffer image. If
   * <code>clear</code> is true, the image is cleared first.
   *
   * @param clear if true, the image is cleared, otherwise the image
   * is unchanged.
   * @return the graphics context for the buffer image
   * @see #getImageGraphics()
   */
  public Graphics2D getImageGraphics(boolean clear) {
    Graphics2D g2d = null;

    if(bimg == null)
      bimg = (BufferedImage) createImage(w,h);
    
    g2d = bimg.createGraphics();
    
    if(clear) {
      g2d.setBackground(getBackground());
      g2d.clearRect(0,0,w,h);
    }
    
    g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                         RenderingHints.VALUE_ANTIALIAS_ON);
    g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                         RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    return g2d;
  }

}