package vcp.animation;

/**
 * <code>FPSAnimation</code> extends <code>Animation</code>
 *
 * @author phil kaasa
 * @version 1.0
 */
public abstract class FPSAnimation extends Animation {
  /** the maximum fps */
  static final int MAX_FPS = 80;

  /** the minimum fps */
  static final int MIN_FPS = 10;

  /** the fps step */
  static final int FPS_STEP = 5;

  /** the period (time required for each frame to maintain the desired
   * FPS) */
  protected long period = 12;

  /** the current target fps */
  protected long fps;

  /** used to track the actual fps */
  protected long actualFPS = 0;

  /** used for timing to ensure the target fps */
  protected long beforeTime;

  /**
   * Creates a new <code>FPSAnimation</code> instance.
   *
   * @param cp the ControlPanel for this FPSAnimation
   * @param fpsa the targe FPS
   * @param trackFPS if true, the actual FPS will be printed to the
   * console every half second.
   */
  public FPSAnimation(ControlPanel cp, long fpsa, boolean trackFPS) {
    fps = fpsa;
    period = (long)(1000.0/fps);

    cPanel = cp;
    cPanel.setAnimation(this);
    //cPanel.addActionListener(this);

    //partArray = createPartArray();

    part = 0;

    if(trackFPS)
      (new Thread() {
          public void run() {
            while(true) {
              try {
                sleep(500);
                System.out.println(actualFPS*2);
              } catch (InterruptedException e) {
                System.out.println(e);
              }
              actualFPS = 0;
            }
            
          }
        }).start();
  }

  /**
   * Creates a new <code>FPSAnimation</code> instance.
   *
   * @param cp the ControlPanel for this FPSAnimation
   * @param fpsa the targe FPS
   */
  public FPSAnimation(ControlPanel cp, long fpsa) {
    this(cp,fpsa,false);
  }
  
  /**
   * Cause animation to sleep for the number of milliseconds required
   * to maintain the target FPS.
   * 
   * @see #sleep(int)
   */
  public void sleep() {
    long timeDiff, sleepTime;

    actualFPS++;
    timeDiff = System.currentTimeMillis() - beforeTime;
    sleepTime = period - timeDiff;
    if(sleepTime <= 0)
      sleepTime = 1;
    super.sleep((int)sleepTime);
    beforeTime = System.currentTimeMillis();
  }

  /**
   * Cause animation to sleep <code>n</code> times. That is, cause
   * <code>n</code> frames to pass without changing the animation.
   *
   * @param n the number of times to sleep
   * @see #sleep()
   */
  public void sleep(int n) {
    long timeDiff, sleepTime;

    for(int i = 0; i < n-1; i++) {
      actualFPS++;
      timeDiff = System.currentTimeMillis() - beforeTime;
      sleepTime = period - timeDiff;
      if(sleepTime <= 0)
        sleepTime = 1;
      try {
        Thread.sleep((int)sleepTime);
      } catch (InterruptedException e) {}
      beforeTime = System.currentTimeMillis();
    }


    sleep();
  }

  public void run() {
    beforeTime = System.currentTimeMillis();
    super.run();
  }

  public void slower() {
    fps -= FPS_STEP;
    
    if(fps < MIN_FPS)
      fps = MIN_FPS;

    period = (int)(1000.0/fps);
  }

  public void faster() {
    fps += FPS_STEP;
    
    if(fps > MAX_FPS)
      fps = MAX_FPS;

    period = (int)(1000.0/fps);
  }

  public double getSpeed() {
    double value = (double)
      (
       100*(((double)(fps-MIN_FPS)/
	     (double)(MAX_FPS-MIN_FPS)))
       );
    
    //0% speed doesn't make much sense
    //if(percent == 0) percent = 1;

    return value;
  }
 
  public void setSpeed(double s) {
    fps = (int)((s/100.0)*(MAX_FPS-MIN_FPS) + MIN_FPS);
    period = (int)(1000.0/fps);
  }

}