package vcp.animation;

/**
 * A semaphore implementation similar in functionality to the
 * Semaphore class included in Java 1.5.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class Semaphore {
  /** the number of available permits */
  private int count;

  /**
   * Creates a new <code>Semaphore</code> with 1 available permit.
   *
   */
  public Semaphore() {
    this(1);
  }

  /**
   * Creates a <code>Semaphore</code> with the given number of
   * permits.
   *
   * @param permits the initial number of permits. (can be negative)
   */
  public Semaphore(int permits) {
    count = permits;
  }

  /**
   * Acquire the the given number of permits, blocking until all are
   * available. If enough permetis are available, this returns
   * immediately after reducing the number of availble permits by the
   * given amount. If an insufficient number of permits are available,
   * blocks until enough are or this thread is interrupted.
   *
   * @param permits the (non-negative) number of permits to acquire
   * @exception InterruptedException if the current thread is
   * interrupted
   * @exception IllegalArgumentException if permits is less than zero
   * @see #acquireUninterruptibly(int)
   */
  public synchronized void acquire(int permits) throws InterruptedException {
    if(permits < 0)
      throw new IllegalArgumentException("permits must be greater than " +
                                         "or equal to 0: " + permits);

    if (count < permits)
      wait();
    
    count -= permits;
  }

  /**
   * <p>Acquire the the given number of permits, blocking until all are
   * available. If enough permetis are available, this returns
   * immediately after reducing the number of availble permits by the
   * given amount. If an insufficient number of permits are available,
   * blocks until enough are.</p>
   *
   * <p>This method differs from <code>acquire()</code> in that it
   * will not cease blocking if the thread is interrupted. If the
   * thread is interrupted while waiting for permits, it will continue
   * to wait. When the thread returns from this method, its interrupt
   * status will be set.</p>
   *
   * @param permits the (non-negative) number of permits to acquire
   * @exception IllegalArgumentException if permits is less than zero
   * @see #acquire(int)
   */
  public synchronized void acquireUninterruptibly(int permits) {
    if(permits < 0)
      throw new IllegalArgumentException("permits must be greater than " +
                                         "or equal to 0: " + permits);

    boolean interrupted = false;

    while(count < permits) {
      try {
        wait();
      } catch(InterruptedException e) {
        interrupted = true;
      }
    }

    count -= permits;

    if(interrupted)
      Thread.currentThread().interrupt();
  }

  /**
   * Acquires one permit.
   *
   * @exception InterruptedException if the current thread is
   * interrupted
   * @see #acquire(int)
   */
  public void acquire() throws InterruptedException {
    acquire(1);
  }

  /**
   * Acquires one permit uninterruptibly.
   *
   * @see #acquireUninterruptibly(int)
   */
  public void acquireUninterruptibly() {
    acquireUninterruptibly(1);
  }

  /**
   * Releases the given number of permits. This increases the number
   * of available permits by that amount.
   *
   * @param permits number of permits to release.
   * @exception IllegalArgumentException if permits is less than zero
   */
  public synchronized void release(int permits) {
    if(permits < 0)
      throw new IllegalArgumentException("permits must be greater than " +
                                         "or equal to 0: " + permits);

    notify();
    
    count += permits;
  }

  /**
   * Releases one permit.
   *
   * @see #release(int)
   */
  public void release() {
    release(1);
  }

  /**
   * Returns the number of available permits.
   *
   * @return the number of available permits
   */
  public synchronized int availablePermits() {
    return count;
  }

}