package vcp.mathnotation;

import java.awt.RenderingHints;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Font;
import java.awt.Color;
import java.awt.image.BufferedImage;
import vcp.utilities.UnicodeUtil;
import vcp.utilities.NoFontFoundException;

/**
 * This represents a mathematical expression. An expression is made up
 * of an array of MathSymbols
 *
 * @author phil kaasa
 * @version 1.0
 */
public class MathExpression {
  /** the array of MathSymbols for this expression */
  protected MathSymbol[] symbol;

  /** the base size of this expression */
  protected float size;

  /** the graphics used to render this expression */
  protected Graphics graphics;

  /** a default space */
  protected MathSpace space;

  public MathExpression(MathSymbol s, float sz) {
    this(new MathSymbol[] {s}, sz);
  }

  public MathExpression(MathSymbol s) {
    this(new MathSymbol[] {s});
  }

  /**
   * Creates a new <code>MathExpression</code> with base size 1.
   *
   * @param s the symbols in the expression
   */
  public MathExpression(MathSymbol[] s) {
    this(s,1);
  }

  /**
   * <p>Creates a new <code>MathExpression</code>.</p>
   *
   * <p>The symbols in the array will be displayed from left to right in
   * ascending order (i.e. index 0 will be the leftmost symbol in the
   * expression). The point size will be the size used for most
   * "normal" text in the expression. Some symbols (e.g. summation)
   * may be quite a bit larger than the given point size.</p>
   *
   * <p>{@link MathSpace#getDefaultSpace Default spaces} will be
   * inserted between all non-{@link MathSpace} symbols in the
   * array. So, if you don't need any custom spacing, you don't need
   * to include any spacing characters. If you do need custom spacing,
   * you only need to include the custom spaces; default spaces will
   * still be inserted where no MathSpace character appears.</p>
   *
   * @param s the symbols in the expression 
   * @param sz the point size of the expression
   */
  public MathExpression(MathSymbol[] s, float sz) {
    symbol = s;
    /*
    size = sz;

    for(int i = 0; i < symbol.length; i++)
      if(symbol[i] != null)
        symbol[i].setSize(size);
    */
    
    space = MathSpace.getDefaultSpace();

    setSize(sz);
    setColor(Color.black);

    graphics = null;
  }

  public void setColor(Color c) {
    for(int i = 0; i < symbol.length; i++)
      if(symbol[i] != null)
        symbol[i].setColor(c);
  }

  /**
   * Set the base size of this expression
   *
   * @param sz the size
   */
  public void setSize(float sz) {
    size = sz;

    for(int i = 0; i < symbol.length; i++)
      if(symbol[i] != null)
        symbol[i].setSize(size);

    space.setSize(sz);
  }

  /**
   * Returns the base size of this expression.
   *
   * @return the base size of this expression
   */
  public float getSize() {
    return size;
  }

  /**
   * Returns the height of the tallest symbol in the expression when
   * displayed using the given graphics object.
   *
   * @param g a <code>Graphics</code> value
   * @return the height of the tallest symbol in the expression
   */
  public double getHeight(Graphics g) {
    double max = 0;

    setGraphics(g);

    for(int i = 0; i < symbol.length; i++) {
      if(symbol[i] != null && symbol[i].getHeight() > max)
        max = symbol[i].getHeight();
    }

    return max;
  }

  /**
   * Returns the width of the expression.
   *
   * The width returned is if it this is displayed using the graphics
   * context set with the last call to {@link #setGraphics(Graphics)}.
   *
   * @return the width of the expression
   */
  public double getWidth() {
    if(graphics == null)
      throw new IllegalStateException();

    return getWidth(graphics);
  }

 
  /**
   * Returns the height of the expression.
   *
   * The height returned is if it this is displayed using the graphics
   * context set with the last call to {@link #setGraphics(Graphics)}.
   *
   * @return the height of the expression
   */  public double getHeight() {
    if(graphics == null)
      throw new IllegalStateException();

    return getHeight(graphics);
  }

  /**
   * Returns the width of the expression when it is displayed using
   * the given graphics object.
   *
   * @param g a <code>Graphics</code> value
   * @return the width of the expression
   */
  public double getWidth(Graphics g) {
    setGraphics(g);

    double w = 0;

    for(int i = 0; i < symbol.length; i++)
      if(symbol[i] != null) {
        w += symbol[i].getWidth();
        if(!(symbol[i] instanceof MathSpace) && 
           (i+1) < symbol.length && 
           !(symbol[i+1] instanceof MathSpace)) {
          w+=space.getWidth();
        }
      }

    return w;// - space;
  }

  /**
   * Set the graphics object for each of the symbols in this
   * expression.
   *
   * @param g a <code>Graphics</code> value
   * @see MathSymbol#setGraphics(Graphics)
   */
  protected void setGraphics(Graphics g) {
    if(graphics != g) {
      space.setGraphics(g); /*= (float)
        g.getFontMetrics(new Font
        ("Serif", Font.PLAIN,(int)size)).stringWidth(" ");*/

      Graphics2D g2d = (Graphics2D) g;
      
      g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                           RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
      g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                           RenderingHints.VALUE_ANTIALIAS_ON);

      for(int i = 0; i < symbol.length; i++)
        if(symbol[i] != null)
          symbol[i].setGraphics(g);

      graphics = g;
    }
  }

  /**
   * Draw this expression with the given Graphics object as the given
   * coordinates.
   *
   * @param g a <code>Graphics</code> value
   * @param x the x position
   * @param y the y position
   */
  public void draw(Graphics g, float x, float y) {
    draw(g,x,y,false);
  }

  /**
   * Draw this expression with the given Graphics object as the given
   * coordinates.
   *
   * @param g a <code>Graphics</code> value
   * @param x the x position
   * @param y the y position
   * @param lines if true, lines will be drawn that are useful for
   * debugging a custom MathSymbol.
   */
  public void draw(Graphics g, float x, float y, boolean lines) {
    setGraphics(g);

    float mid = y - (float)getHeight(g)/2f;
    //float base = y - (float)getHeight(g);

    for(int i = 0; i < symbol.length; i++) {
      //System.out.println(i + ": x = " + x);
      if(lines)
        ((Graphics2D) g).draw(new java.awt.geom.Line2D.Float(x,0f,x,y));
      if(symbol[i] != null) {
        symbol[i].draw(x,mid + (float)symbol[i].getHeight()/2f);
        //symbol[i].draw(x,y);
        x+=symbol[i].getWidth();
          
        if(lines)
          ((Graphics2D) g).draw(new java.awt.geom.Line2D.Float(x,0f,x,y));
        
        if(!(symbol[i] instanceof MathSpace) && 
           (i+1) < symbol.length && 
           !(symbol[i+1] instanceof MathSpace)) {
          space.draw(x,mid + (float)space.getHeight()/2f);
          x+=space.getWidth();
        }
        //x += space;
      }
    }
    
    //for testing
    if(lines) {
      ((Graphics2D) g).draw(new java.awt.geom.Line2D.Float(0f,mid,(float) g.getClipBounds().getWidth(),mid));
      ((Graphics2D) g).draw(new java.awt.geom.Line2D.Float(0f,y,(float) g.getClipBounds().getWidth(),y));
      ((Graphics2D) g).draw(new java.awt.geom.Line2D.Float(0f,y-(float)getHeight(g),(float) g.getClipBounds().getWidth(),y-(float)getHeight(g)));
    }
  }

  /**
   * Returns a <code>BufferedImage</code> of this
   * <code>MathExpression</code>. The background color of the image is
   * as passed.
   *
   * @param margin the number of pixels between the expression and the
   * edge of the image.
   * @return a <code>BufferedImage</code> of this
   * <code>MathExpression</code>
   */
  public BufferedImage toImage(int margin) {
    BufferedImage img = new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = (Graphics2D) img.getGraphics();
    int w = (int)getWidth(g);
    int h = (int)getHeight(g);
    img = new BufferedImage(w+(2*margin), h+(2*margin),
                            BufferedImage.TYPE_INT_ARGB);
    g.dispose();
    g = (Graphics2D) img.getGraphics();
    g.setBackground(new Color(0,0,0,0));
    g.clearRect(0,0,w+(2*margin),h+(2*margin));
    draw(g,margin,h+margin);
    g.dispose();

    return img;
  }

}