package vcp.mathnotation;

import java.awt.Font;
import java.awt.Graphics;

/**
 * The class is used for spacing between symbols in a
 * {@link MathExpression}.
 *
 * @author phil kaasa
 * @version 1.0
 */
public abstract class MathSpace extends MathSymbol {
  /** the point size of this space */
  protected float size;

  /**
   * Returns the width of this space.
   *
   * @return the width of this space.
   */
  public abstract double getWidth();

  /**
   * A <code>MathSpace</code> which has no width.
   */
  public static MathSpace noSpace = new MathSpace() {
      public double getWidth() {
        return 0;
      }   
    };

  /**
   * The width of this space is equal to the width of one space using
   * the Serif logical font at the current size.
   */
  protected static class DefaultSpace extends MathSpace {
    /** the width of the space */
    float space;
    
    public DefaultSpace() {
      super();
    }

    public void setGraphics(Graphics g) {
      space = (float)
        g.getFontMetrics(new Font
                         ("Serif", Font.PLAIN,(int)size)).stringWidth(" ");
    }
    
    public double getWidth() {
      return space;
    }
  }

  /**
   * The width of this space is relative to a default space.
   */
  protected static class RelativeSpace extends DefaultSpace {
    float ratio;

    /**
     * Creates a new <code>RelativeSpace</code>.
     *
     * @param r the ratio of this space to a default space.
     */
    public RelativeSpace(float r) {
      ratio = r;
    }

    public double getWidth() {
      return super.getWidth() * ratio;
    }    
  }

  /**
   * The width of this space is an absolute pixel value.
   */
  protected static class AbsoluteSpace extends MathSpace {
    float space;

    /**
     * Creates a new <code>AbsoluteSpace</code>.
     *
     * @param s the width of this space
     */
    public AbsoluteSpace(float s) {
      space = s;
    }

    public double getWidth() {
      return space;
    }
  }

  
  /**
   * Creates a new <code>MathSpace</code> instance.
   */
  public MathSpace() {
    super();
  }

  /**
   * <p>This method does nothing. A space shouldn't need to draw anything
   * to the screen.</p>
   *
   * Documentation from <code>MathSymbol</code>:<br>
   * {@inheritDoc}
   */
  public void setGraphics(Graphics g) {
    
  }

  /**
   * <p>This method does nothing. A space shouldn't need to draw anything
   * to the screen.</p>
   *
   * Documentation from <code>MathSymbol</code>:<br>
   * {@inheritDoc}
   */
  public void draw(float x, float y) {

  }

  /**
   * Set the size of this MathSpace
   *
   * @param s a point size value
   * @see #getSize()
   */
  public void setSize(float s) {
    size = s;
  }
  
  /**
   * Returns the point size of this MathSpace.
   *
   * @return the point size of this MathSpace
   * @see #setSize(float)
   */
  public float getSize() {
    return size;
  }
  
  /**
   * Returns 0. Most spaces shouldn't have any height.
   *
   * @return 0
   */
  public double getHeight() {
    return 0;
  }  

  /**
   * Returns a default space. The width of this space is equal to the
   * width of one space using the Serif logical font at the current
   * size.
   *
   * @return a default space
   */
  public static MathSpace getDefaultSpace() {
    return new DefaultSpace();
  }

  /**
   * Returns a space with 0 width.
   *
   * @return a space with 0 width
   */
  public static MathSpace getNoSpace() {
    return noSpace;
  }

  /**
   * Returns a space with size relative to a default space.
   *
   * @param ratio the ratio between this space and a default space. A
   * ratio of 2 would give a space twice as wide as a default space.
   * @return a space with size relative to a default space
   */
  public static MathSpace getRelativeSpace(float ratio) {
    return new RelativeSpace(ratio);
  }

  /**
   * Returns a space with an absolute pixel width.
   *
   * @param space the width of this space
   * @return a space with an absolute pixel width
   */
  public static MathSpace getAbsoluteSpace(float space) {
    return new AbsoluteSpace(space);
  }
}