package vcp.mathnotation;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Font;
import java.awt.font.GlyphVector;
import java.awt.geom.Arc2D;
import java.awt.geom.Line2D;

import vcp.utilities.UnicodeUtil;
import vcp.utilities.NoFontFoundException;

/**
 * This class represents a MathSymbol which is used in MathExpressions.
 *
 * @author phil kaasa
 * @version 1.0
 * @see <code>MathExpression</code>
 */
public abstract class MathSymbol {
  /** the current graphics object to use for rendering */
  protected Graphics2D g2d;

  /** the color of the MathSymbol */
  protected Color color = Color.black;

  /**
   * Set the graphics object that will display this MathSymbol.
   *
   * @param g a <code>Graphics</code> value
   */
  protected void setGraphics(Graphics g) {
    g2d = (Graphics2D) g;
  }

  /**
   * Draw this MathSymbol at the specified coordinates. The
   * coordinates specify the location lower left corner of the symbol.
   *
   * @param x the x position
   * @param y the y position
   * @exception IllegalStateException if the graphics object is not
   * yet set.
   */
  public abstract void draw(float x, float y);

  /**
   * Return the width of the MathSymbol. <em>This method must be
   * overriden.</em>
   *
   * @return the width of the MathSymbol
   * @exception IllegalStateException if the graphics object is not
   * yet set.
   */
  protected double getWidth() {
    if(g2d == null)
      throw new IllegalStateException
        ("getWidth cannot be called before setGraphics is called");

    return 0;
  }

  /**
   * Return the height of the MathSymbol. <em>This method must be
   * overriden.</em>
   *
   * @return the height of the MathSymbol
   * @exception IllegalStateException if the graphics object is not
   * yet set.
   */
  protected double getHeight() {
    if(g2d == null)
      throw new IllegalStateException
        ("getHeight cannot be called before setGraphics is called");

    return 0;
  }

  /**
   * Creates a new <code>MathSymbol</code> instance.
   */
  public MathSymbol() {

  }

  /**
   * This class has methods to get the width and height of the actual
   * text.
   */
  protected class ExtendedGlyphVector {
    GlyphVector glyph;
    String string;
    Font font;

    /**
     * Creates a new <code>ExtendedGlyphVector</code> instance. The
     * graphics object of this MathSymbol must be set before an
     * ExtendedGlyphVector can be created.
     *
     * @param s String the <code>ExtendedGlyphVector</code> should display
     * @param f The <code>Font</code> to display the String with
     * @throws IllegalStateException if {@link #setGraphics(Graphics)}
     * has not yet been called.
     */
    public ExtendedGlyphVector(String s, Font f) {
      if(g2d == null)
        throw new IllegalStateException("graphics object is not set");

      glyph = f.createGlyphVector
        (g2d.getFontRenderContext(), s);

      font = f;
      string = s;
    }

    /**
     * Returns the width of the text according to font metrics
     * (instead of actual visible width)
     *
     * @return  the width of the text
     */
    public double getWidth() {
      return g2d.getFontMetrics(font).getStringBounds(string,g2d).getWidth();
    }
 
    /**
     * Returns the height of the text according to font metrics
     * (instead of actual visible height)
     *
     * @return  the height of the text
     */
    public double getHeight() {
      return g2d.getFontMetrics(font).getStringBounds(string,g2d).getHeight();
    }

    protected java.awt.Rectangle getGlyphBounds() {
      return getGlyphBounds(0,0);
    }

    protected java.awt.Rectangle getGlyphBounds(float x, float y) {
      //OS X java returns strange values for getPixelBounds, so we
      //need to use different methods there.
      if(System.getProperty("os.name").startsWith("Mac OS X"))
        return glyph.getOutline(x,y).getBounds();
      else
        return glyph.getPixelBounds(null,x,y);
    }

    /**
     * Returns the height of the GlyphVector according to visual
     * bounds
     *
     * @return the height of the GlyphVector
     */
    protected double getGlyphHeight() {
      return getGlyphBounds().getHeight();
    }

    /**
     * Returns the width of the GlyphVector according to visual bounds
     *
     * @return the width of the GlyphVector
     */
    protected double getGlyphWidth() {
      return getGlyphBounds().getWidth();
    }
    
  }


  /**
   * Returns a MathSymbol which is a summation
   *
   * @param indexS the index of the summation (e.g. "i = 0")
   * @param limitS the limit of the summation (e.g. n)
   * @param expression a MathExpression which is the expression to be
   * summed.
   * @param alt the alternate text to be displayed if appropriate
   * characters cannot be found.
   * @return a MathSymbol which is a summation
   */
  public static MathSymbol getSummationSymbol(String indexS, 
                                              String limitS, 
                                              MathExpression expression,
                                              String alt) {
    return getSummationSymbol(getTextSymbol(indexS).toMathExpression(),
                              getTextSymbol(limitS).toMathExpression(), 
                              expression,
                              alt);
  }  

  /**
   * Returns a MathSymbol which is a summation
   *
   * @param indexS the index of the summation (e.g. "i = 0")
   * @param limitS the limit of the summation (e.g. n)
   * @param expression a MathSymbol which is the expression to be
   * summed.
   * @param alt the alternate text to be displayed if appropriate
   * characters cannot be found.
   * @return a MathSymbol which is a summation
   */
  public static MathSymbol getSummationSymbol(String indexS, 
                                              String limitS, 
                                              MathSymbol expression,
                                              String alt) {
    return getSummationSymbol(getTextSymbol(indexS).toMathExpression(),
                              getTextSymbol(limitS).toMathExpression(), 
                              expression.toMathExpression(),
                              alt);
  }

  /**
   * Returns a MathSymbol which is a summation.
   *
   * @param indexS the index of the summation (e.g. "i = 0")
   * @param limitS the limit of the summation (e.g. n)
   * @param expression the expression to be summed.
   * @param alt the alternate text to be displayed if appropriate
   * characters cannot be found.
   * @return a MathSymbol which is a summation
   */
  public static MathSymbol getSummationSymbol(String indexS, 
                                              String limitS, 
                                              String expression,
                                              String alt) {
    return getSummationSymbol(getTextSymbol(indexS).toMathExpression(),
                              getTextSymbol(limitS).toMathExpression(), 
                              getTextSymbol(expression).toMathExpression(),
                              alt);
  }

  /**
   * Returns a MathSymbol which is a summation.
   *
   * @param indexS the index of the summation (e.g. "i = 0")
   * @param limitS the limit of the summation (e.g. n)
   * @param expression a MathExpression which is the expression to be
   * summed.
   * @param alt the alternate text to be displayed if appropriate
   * characters cannot be found.
   * @return a MathSymbol which is a summation
   */
  public static MathSymbol getSummationSymbol(MathExpression indexS, 
                                              MathExpression limitS, 
                                              MathExpression expression,
                                              String alt) {
    class SummationSymbol extends MathSymbol {
      //String indexS;
      //String limitS;
      MathExpression expression;
      
      String sigmaStr = "\u03a3";

      ExtendedGlyphVector sigma;
      MathExpression index;
      MathExpression limit;

      MathSpace sp;
      
      Font f;
   
      public SummationSymbol(MathExpression i, MathExpression l, 
                             MathExpression e)
        throws NoFontFoundException {

        super();

        index = i;
        limit = l;
        expression = e;

        try {
          f = UnicodeUtil.getDisplayFont(sigmaStr,UnicodeUtil.SERIF);
        } catch (NoFontFoundException nffe) {
          sigmaStr = "\u2211";
          f = UnicodeUtil.getDisplayFont(sigmaStr,UnicodeUtil.SERIF);          
        }

        sp = MathSpace.getRelativeSpace(.5f);

        g2d = null;
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        sigma = new ExtendedGlyphVector(sigmaStr,scaleFont(f,155)); 
      
        expression.setGraphics(g);
        index.setGraphics(g);
        limit.setGraphics(g);
        sp.setGraphics(g);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        //y -= getGlyphHeight(index.glyph) + 4;
        y -= index.getHeight(g) + 4;

        double width;

        if(sigma.getWidth() > index.getWidth(g)) {
          width = sigma.getWidth();

          g.drawGlyphVector(sigma.glyph,x,y);
          
          index.draw(g,
                     (float)(x+sigma.getWidth()/2.0 - index.getWidth(g)/2.0),
                     (float)(y + index.getHeight(g) + 4));

          /*
          g.drawGlyphVector
            (index.glyph,
             (float)(x+sigma.getWidth()/2.0 - index.getWidth()/2.0),
             (float)(y + index.getGlyphHeight() + 4));
          */

          expression.draw
            (g,
             (float)(sigma.glyph.getOutline(x,y).getBounds().getX() + sigma.getGlyphWidth() + 2),
             (float)(y - sigma.getGlyphHeight()/2.0 + 
                     expression.getHeight(g)/2.0));
          
        } else {
          width = index.getWidth(g);

          index.draw(g,
                     x,(float)(y + index.getHeight(g) + 4));

          /*
          g.drawGlyphVector(index.glyph,x,
                              (float)(y + index.getGlyphHeight() + 4));
          */

          g.drawGlyphVector
            (sigma.glyph,
             (float)(x+index.getWidth(g)/2.0 - sigma.getWidth()/2.0),
             y);

          expression.draw
            (g,
             (float)(sigma.glyph.getOutline((float)(x+index.getWidth(g)/2.0 - 
                                                    sigma.getWidth()/2.0),y).
                     getBounds().getX() + sigma.getGlyphWidth() + sp.getWidth()),
             (float)(y - sigma.getGlyphHeight()/2.0 + 
                     expression.getHeight(g)/2.0));
        }
 
        limit.draw(g,
                   (float)(x + width/2.0 - limit.getWidth(g)/2.0),
                   (float)(y - sigma.getGlyphHeight() - 4));

        /*
        g.drawGlyphVector
          (limit.glyph,
           (float)(x + width/2.0 - limit.getWidth()/2.0),
           (float)(y - sigma.getGlyphHeight() - 4));
        */
        g.dispose();
      }
        
      protected double getWidth() {
        super.getWidth();
        
        return (sigma.getWidth() > index.getWidth(g2d) ? 
                sigma.getWidth() : index.getWidth(g2d)/2.0 + 
                sigma.getWidth()/2.0) + 
          expression.getWidth(g2d) + sp.getWidth() - 
          sigma.glyph.getOutline(0f,0f).getBounds().getX();
      }

      protected double getHeight() {
        super.getHeight();

        return getGlyphHeight(sigma.glyph) + index.getHeight(g2d) + 
          limit.getHeight(g2d) + 8;
      }

      public void setSize(float s) {
        f = f.deriveFont(s);
        expression.setSize(s);
        index.setSize(.5f*s);
        limit.setSize(.5f*s);
        sp.setSize(s);
      }

      public float getSize() {
        return f.getSize();
      }

      public void setColor(Color c) {
        super.setColor(c);

        expression.setColor(c);
        index.setColor(c);
        limit.setColor(c);
      }
    }

    try {
      return new SummationSymbol(indexS, limitS, expression);
    } catch (NoFontFoundException e) {
      return getTextSymbol("[" + alt + "]");
    }
  }


 /**
   * Returns a MathSymbol which is an instance of binomial coefficient
   * notation. Produces the notation which is equivalent to C(n,r).
   *
   * @param n the number of things to choose from
   * @param r the number of things to take from n
   * @return a MathSymbol which is an instance of binomial coefficient
   * notation
   */
  public static MathSymbol getBinomialCoefficient(String n,
                                                  String r) {
    return getBinomialCoefficient
      (getTextSymbol(n).toMathExpression(),
       getTextSymbol(r).toMathExpression());
  }

  /**
   * Returns a MathSymbol which is an instance of binomial coefficient
   * notation. Produces the notation which is equivalent to C(n,r).
   *
   * @param n the number of things to choose from
   * @param r the number of things to take from n
   * @return a MathSymbol which is an instance of binomial coefficient
   * notation
   */
  public static MathSymbol getBinomialCoefficient(MathExpression n,
                                                  MathExpression r) {
    
    class BinomialCoefficient extends MathSymbol {
      static final float arcRatio = .52f;

      //String topS;
      //String botS;

      double w;
      double h;
      double vgap;
      double hgap;
      int strokeWidth;

      double arcChord;
      double arcWidth;

      //ExtendedGlyphVector top;
      //ExtendedGlyphVector bot;
      
      MathExpression top;
      MathExpression bot;

      float size;

      //Font f;

      public BinomialCoefficient(MathExpression t, MathExpression b)
        throws NoFontFoundException {
        super();
        
        top = t;
        bot = b;

        //f = UnicodeUtil.getAlphaNumberFont();
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        //top = new ExtendedGlyphVector(topS, scaleFont(f,50));
        //bot = new ExtendedGlyphVector(botS, scaleFont(f,50));
        
        top.setGraphics(g);
        bot.setGraphics(g);

        w = (top.getWidth() > bot.getWidth()) ?
          top.getWidth() : bot.getWidth();
        h = bot.getHeight()+top.getHeight();
        vgap = h/3.5;
        //hgap = w*1.25;
        hgap = 0;

        Arc2D.Double arc = new Arc2D.Double
          (0,//0 + (float)Math.ceil(strokeWidth/2),
           0,//0 - h - 2 * vgap,
           arcRatio * (h + 2 * vgap), h + 2 * vgap,
           125, 110, Arc2D.OPEN);

        arcChord = arc.getEndPoint().getY() -  arc.getStartPoint().getY();
        arcWidth = arc.getEndPoint().getX() + 3;
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        strokeWidth = (int)Math.ceil((double)getSize()/30.0);

        g.setStroke(new BasicStroke(strokeWidth));
      
        double arcw = arcRatio * (h + 2 * vgap);

        Arc2D.Double arc = new Arc2D.Double
          (x + (float)Math.ceil(strokeWidth/2),
           y - h - 2 * vgap,
           arcw, h + 2 * vgap,
           125, 110, Arc2D.OPEN);

        double arcgap = arc.getStartPoint().getY() - arc.getY();

        y += arcgap;


        g.draw(new Arc2D.Double
                 (x + (float)Math.ceil(strokeWidth/2),
                  y - h - 2 * vgap,
                  arcw, h + 2 * vgap,
                  125, 110, Arc2D.OPEN));
                  //0, 360, Arc2D.OPEN));
                  
        
        g.draw(new Arc2D.Double
                 (x - (float)Math.ceil(strokeWidth/2) + getWidth() - arcw,
                  y - h - 2 * vgap,
                  arcw, h + 2 * vgap,
                  305, 110, Arc2D.OPEN));
                  //0, 360, Arc2D.OPEN));


        /*
        g.drawGlyphVector(bot.glyph,
                            (float)(x + (w - bot.getWidth())/2.0 + arcWidth),
                            (float)(y-vgap));
        g.drawGlyphVector(top.glyph,
                            (float)(x + (w - top.getWidth())/2.0 + arcWidth),
                            (float)(y - h));
        */

        bot.draw(g,(float)(x + (w - bot.getWidth())/2.0 + arcWidth),
                 (float)(y-vgap));
        top.draw(g,(float)(x + (w - top.getWidth())/2.0 + arcWidth),
                 (float)(y - h));

        g.dispose();
      }

      protected double getWidth() {
        super.getWidth();

        return w + (2 * arcWidth) + Math.ceil(strokeWidth/2);
      }
      
      protected double getHeight() {
        super.getHeight();

        return arcChord;
      }

      public void setSize(float s) {
        //f = f.deriveFont(s);
        top.setSize(s * .5f);
        bot.setSize(s * .5f);
        size = s;
      }

      public float getSize() {
        //return f.getSize();
        return size;
      }

      public void setColor(Color c) {
        super.setColor(c);

        top.setColor(c);
        bot.setColor(c);
      }
    }   

    try {
      return new BinomialCoefficient(n,r);
    } catch (NoFontFoundException e) {
      return getTextSymbol("[C(" + n + ", " + r + ")]");
    }
  }

  /**
   * Returns a MathSymbol which is an expression with a superscript.
   *
   * @param base expression to have subscript
   * @param superscript superscript of the expression
   * @return a MathSymbol which is an expression with a superscript
   * @see #getSubscript(String,String)
   */
  public static MathSymbol getSuperscript(String base,
                                          String superscript) {
    return getSuperscript(getTextSymbol(base).toMathExpression(),
                          getTextSymbol(superscript).toMathExpression());
  }

  /**
   * Returns a MathSymbol which is an expression with a superscript.
   *
   * @param base expression to have subscript
   * @param superscript superscript of the expression
   * @return a MathSymbol which is an expression with a superscript
   * @see #getSubscript(String,String)
   */
  public static MathSymbol getSuperscript(MathExpression base,
                                          String superscript) {
    return getSuperscript(base,
                          getTextSymbol(superscript).toMathExpression());
  }

  /**
   * Returns a MathSymbol which is an expression with a superscript.
   *
   * @param base expression to have subscript
   * @param superscript superscript of the expression
   * @return a MathSymbol which is an expression with a superscript 
   * @see #getSubscript(MathExpression,MathExpression)
   */
  public static MathSymbol getSuperscript(MathExpression base, 
                                          MathExpression superscript) {
    class SuperScript extends MathSymbol {
      MathExpression base;
      MathExpression superscript;

      //String baseS;
      //String superS;

      float hgap;


      public SuperScript(MathExpression b, MathExpression s) {
        super();

        base = b;
        superscript = s;
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        base.setGraphics(g);
        superscript.setGraphics(g);

        hgap = (float) Math.ceil(getSize()/25.0);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        base.draw(g, x,y);
        superscript.draw(g,
                         (float)(x + base.getWidth() + hgap),
                         (float)(y - base.getHeight() + 
                                 superscript.getHeight() * .8));

        g.dispose();
        
      }

      protected double getWidth() {
        super.getWidth();

        return base.getWidth() + superscript.getWidth() + hgap;
      }

      protected double getHeight() {
        super.getHeight();

        return base.getHeight() + superscript.getHeight() * 0.2;
      }      

      public void setSize(float s) {
        base.setSize(s);
        superscript.setSize(s*.5f);
      }
    
      public float getSize() {
        return base.getSize();
      }
      public void setColor(Color c) {
        super.setColor(c);

        base.setColor(c);
        superscript.setColor(c);
      }
    }

    //try {
    return new SuperScript(base,superscript);
      //} catch (NoFontFoundException e) {
      //return getTextSymbol("[" + base + "^" + superscript + "]");
      //}
      
  }

  /**
   * Returns a MathSymbol which is an expression with a subscript.
   *
   * @param base expression to have subscript
   * @param subscript subscript of the expression
   * @return a MathSymbol which is an expression with a subscript
   * @see #getSuperscript(String,String)
   */
  public static MathSymbol getSubscript(String base,
                                        String subscript) {
    return getSubscript(getTextSymbol(base).toMathExpression(),
                        getTextSymbol(subscript).toMathExpression());
  }

  /**
   * Returns a MathSymbol which is an expression with a subscript.
   *
   * @param base expression to have subscript
   * @param subscript subscript of the expression
   * @return a MathSymbol which is an expression with a subscript
   * @see #getSuperscript(MathExpression,MathExpression)
   */
  public static MathSymbol getSubscript(MathExpression base, 
                                        MathExpression subscript) {
    class SubScript extends MathSymbol {
      MathExpression base;
      MathExpression subscript;

      //String baseS;
      //String superS;

      float hgap;


      public SubScript(MathExpression b, MathExpression s) {
        super();

        base = b;
        subscript = s;
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        base.setGraphics(g);
        subscript.setGraphics(g);

        hgap = (float) Math.ceil(getSize()/50.0);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        base.draw(g, x,y);
        subscript.draw(g,
                       (float)(x + base.getWidth() + hgap),
                       (float)(y));

        g.dispose();
        
      }

      protected double getWidth() {
        super.getWidth();

        return base.getWidth() + subscript.getWidth() + hgap;
      }

      protected double getHeight() {
        super.getHeight();

        return base.getHeight();
      }      

      public void setSize(float s) {
        base.setSize(s);
        subscript.setSize(s*.5f);

      }
    
      public float getSize() {
        return base.getSize();
      }

      public void setColor(Color c) {
        super.setColor(c);

        base.setColor(c);
        subscript.setColor(c);
      }
    }

    //try {
    return new SubScript(base,subscript);
      //} catch (NoFontFoundException e) {
      //return getTextSymbol("[" + base + "^" + subscript + "]");
      //}
      
  }

  /**
   * Returns an equals symbol. A call to
   * <code>getTextSymbol("=")</code> calls this method.
   *
   * @return an equals symbol
   */
  public static MathSymbol getEqualsSymbol() {
    class Equals extends MathSymbol {
      ExtendedGlyphVector eq;

      Font f;

      public Equals() {
        super();

        f = new Font("serif",Font.PLAIN,1);
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        eq = new ExtendedGlyphVector("=",f);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        java.awt.Rectangle r = eq.getGlyphBounds(x,y);
        
        float newy = y + (y - r.y - r.height);

        g.drawGlyphVector(eq.glyph,x,newy);
      }

      public double getHeight() {
        super.getHeight();

        return eq.getGlyphHeight();
      }

      public double getWidth() {
        super.getWidth();

        return eq.getWidth();
      }

      public void setSize(float s) {
        f = f.deriveFont(s);
      }

      public float getSize() {
        return f.getSize();
      }
    }

    return new Equals();
  }

  /**
   * Returns a MathSymbol which is an expression inside matching floor
   * symbols.
   *
   * @param e the {@link MathExpression} to put in a floor symbol
   * @return a MathSymbol which is an expression inside matching floor
   * symbols
   */
  public static MathSymbol getFloorSymbol(MathExpression e) {
    class FloorSymbol extends MathSymbol {
      MathExpression expr;

      int strokeWidth;
      float width;

      public FloorSymbol(MathExpression e) {
        expr = e;
        strokeWidth = (int)Math.ceil((double)expr.getSize()/30.0);
        width = (float)(strokeWidth * 5);
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        expr.setGraphics(g);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        width = (float)(strokeWidth * 5);

        g.setStroke(new BasicStroke(strokeWidth));

        g.draw(new Line2D.Float(x,y,
                                  x,(float)(y-expr.getHeight())));
        g.draw(new Line2D.Float(x,y,x+width,y));

        float add = (float)(expr.getWidth() + width*2 + strokeWidth*4);

        g.draw(new Line2D.Float((float)(x+add),y,
                                  (float)(x+add),
                                  (float)(y-expr.getHeight())));
        g.draw(new Line2D.Float((float)(x+add),y,
                                  (float)(x+add-width),y));

        expr.draw(g,
                  (float)(x+width+strokeWidth*2),
                  (float)(y));

        g.dispose();
      }

      protected double getWidth() {
        super.getWidth();

        return expr.getWidth() + width*2 + strokeWidth*4;
      }

      protected double getHeight() {
        super.getHeight();

        return expr.getHeight();
      }      

      public void setSize(float s) {
        expr.setSize(s);
        strokeWidth = (int)Math.ceil((double)s/30.0);
        width = (float)(strokeWidth * 5);
      }
    
      public float getSize() {
        return expr.getSize();
      }
      
      public void setColor(Color c) {
        super.setColor(c);

        expr.setColor(c);
      }
    }
    
    return new FloorSymbol(e);
  }

  /**
   * Returns the passed expression enclosed in square brackets.
   *
   * @param e a <code>MathExpression</code> value
   * @return the passed expression enclosed in square brackets.
   */
  public static MathSymbol getBrackets(MathExpression e) {
    class Bracket extends MathSymbol {
      MathExpression expr;

      int strokeWidth;
      float width;

      public Bracket(MathExpression e) {
        expr = e;
        strokeWidth = (int)Math.ceil((double)expr.getSize()/30.0);
        width = (float)(strokeWidth * 5);
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        expr.setGraphics(g);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        width = (float)(strokeWidth * 5);

        g.setStroke(new BasicStroke(strokeWidth));

        g.draw(new Line2D.Float(x,y,
                                  x,(float)(y-expr.getHeight())));
        g.draw(new Line2D.Float(x,y,x+width,y));
        g.draw(new Line2D.Float(x,(float)(y-expr.getHeight()),
                                x+width,(float)(y-expr.getHeight())));

        float add = (float)(expr.getWidth() + width*2 + strokeWidth*4);

        g.draw(new Line2D.Float((float)(x+add),y,
                                (float)(x+add),
                                (float)(y-expr.getHeight())));
        g.draw(new Line2D.Float((float)(x+add),y,
                                (float)(x+add-width),y));
        g.draw(new Line2D.Float((float)(x+add),(float)(y-expr.getHeight()),
                                (float)(x+add-width),
                                (float)(y-expr.getHeight())));

        expr.draw(g,
                  (float)(x+width+strokeWidth*2),
                  (float)(y));

        g.dispose();
      }

      protected double getWidth() {
        super.getWidth();

        return expr.getWidth() + width*2 + strokeWidth*4;
      }

      protected double getHeight() {
        super.getHeight();

        return expr.getHeight();
      }      

      public void setSize(float s) {
        expr.setSize(s);
        strokeWidth = (int)Math.ceil((double)s/30.0);
        width = (float)(strokeWidth * 5);
      }
    
      public float getSize() {
        return expr.getSize();
      }
      
      public void setColor(Color c) {
        super.setColor(c);

        expr.setColor(c);
      }      
    }

    return new Bracket(e);
  }

  /**
   * Returns a MathSymbol which is the passed MathSymbol enclosed
   * in parentheses.
   *
   * @param s The String to be enclosed
   * @return the passed MathSymbol enclosed in parentheses
   */  
  public static MathSymbol getParens(String s) {
    return getParens(MathSymbol.getTextSymbol(s));
  }

  /**
   * Returns a MathSymbol which is the passed MathSymbol enclosed
   * in parentheses.
   *
   * @param e The MathSymbol to be enclosed
   * @return the passed MathSymbol enclosed in parentheses
   */  
  public static MathSymbol getParens(MathSymbol e) {
    return getParens(e.toMathExpression());
  }

  /**
   * Returns a MathSymbol which is the passed MathExpression enclosed
   * in parentheses.
   *
   * @param e The MathExpression to be enclosed
   * @return the passed MathExpression enclosed in parentheses
   */
  public static MathSymbol getParens(MathExpression e) {
    class Parens extends MathSymbol {
      MathExpression expr;

      static final float arcRatio = .52f;
      
      double w;
      double h;
      double vgap;
      double hgap;
      int strokeWidth;

      double arcChord;
      double arcWidth;
      public Parens(MathExpression e) {
        expr = e;
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        expr.setGraphics(g);

        w = expr.getWidth();
        h = expr.getHeight();
        //vgap = 0;
        vgap = h/6;
        //hgap = w*1.25;
        hgap = 0;

        Arc2D.Double arc = new Arc2D.Double
          (0,//0 + (float)Math.ceil(strokeWidth/2),
           0,//0 - h - 2 * vgap,
           arcRatio * (h + 2 * vgap), h + 2 * vgap,
           125, 110, Arc2D.OPEN);

        arcChord = arc.getEndPoint().getY() -  arc.getStartPoint().getY();
        arcWidth = arc.getEndPoint().getX() + 3;
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        strokeWidth = (int)Math.ceil((double)getSize()/30.0);

        g.setStroke(new BasicStroke(strokeWidth));
      
        double arcw = arcRatio * (h + 2 * vgap);

        Arc2D.Double arc = new Arc2D.Double
          (x + (float)Math.ceil(strokeWidth/2),
           y - h - 2 * vgap,
           arcw, h + 2 * vgap,
           125, 110, Arc2D.OPEN);

        double arcgap = arc.getStartPoint().getY() - arc.getY();

        y += arcgap;


        g.draw(new Arc2D.Double
                 (x + (float)Math.ceil(strokeWidth/2),
                  y - h - 2 * vgap,
                  arcw, h + 2 * vgap,
                  125, 110, Arc2D.OPEN));
        
        g.draw(new Arc2D.Double
                 (x - (float)Math.ceil(strokeWidth/2) + getWidth() - arcw,
                  y - h - 2 * vgap,
                  arcw, h + 2 * vgap,
                  305, 110, Arc2D.OPEN));

        expr.draw(g,
                  (float)(x + arcWidth),
                  (float)(y-vgap));

        g.dispose();
      }

      public double getWidth() {
        super.getWidth();

        return w + 2 * arcWidth + Math.ceil(strokeWidth/2);
      }

      public double getHeight() {
        super.getHeight();

        return expr.getHeight();
      }

      public float getSize() {
        return expr.getSize();
      }

      public void setSize(float s) {
        expr.setSize(s);
      }

      public void setColor(Color c) {
        super.setColor(c);

        expr.setColor(c);
      }
    }

    return new Parens(e);
  }
  

  /**
   * Returns a MathSymbol which is just text. This can be used for
   * simple notation e.g. <em>x</em> + <em>y</em> = <em>z</em>. All
   * Letter characters will be italicized.
   *
   * @param text the text to display
   * @return a MathSymbol which is just text
   */
  public static MathSymbol getTextSymbol(String text) {
    return getTextSymbol(text,Font.ITALIC);
  }

  /**
   * Returns a MathSymbol which is just text. This can be used for
   * simple notation e.g. x + y = z.
   *
   * @param text the text to display
   * @param letterStyle the font style for letters
   * (<code>Font.PLAIN</code>, <code>Font.ITALIC</code>,
   * <code>Font.BOLD</code>)
   * @return a MathSymbol which is just text
   */
  public static MathSymbol getTextSymbol(String text, int letterStyle) {
    class TextSymbol extends MathSymbol {
      String[] text;
      ExtendedGlyphVector[] t;

      Font[] f;

      Font fp;
      Font fi;

      String tx;

      TextSection head = null;
      TextSection tail = null;

      class TextSection {
        ExtendedGlyphVector extendedGlyph;
        
        TextSection next;
        
        String string;
        
        Font font;
        
        public TextSection(String s, Font f) {
          this(s,f,null);
        }
        
        public TextSection(String s, Font f, TextSection n) {
          string = s;
          font = f;
          next = n;
        }
        
        public void createExtendedGlyphVector() {
          extendedGlyph = new ExtendedGlyphVector(string,font);
        }

        public void draw(Graphics2D g, float x, float y) {
          g.drawGlyphVector(extendedGlyph.glyph,x,y);
        }
      }
      

      public TextSymbol(String txt, int letterStyle) {
        super();

        tx = txt;

        //will match any group of letter characters
        String expr = "\\p{L}+";

        Matcher m = Pattern.compile(expr).matcher(txt);

        int last = 0;
        String toAdd = null;

        //separate input string into to letter and non-letter
        //sections.
        while(m.find()) {
          //a non-letter section
          if(m.start() > last)
            try {        
              toAdd = txt.substring(last,m.start());
              addSection(toAdd,
                         UnicodeUtil.getDisplayFont
                         (toAdd,
                          UnicodeUtil.SERIF).deriveFont(Font.PLAIN));
            } catch (NoFontFoundException e) {
              Font f;
              if(e.getFontList().length > 0)
                f = e.getFontList()[e.getBestFont()];
              else
                f = new Font("serif",Font.PLAIN,1);
              addSection(UnicodeUtil.fixString(toAdd,f),
                         f.deriveFont(Font.PLAIN));
            }
          
          //a letter section
          try {              
            toAdd = m.group();
            addSection(toAdd,
                       UnicodeUtil.getDisplayFont
                       (toAdd,
                        UnicodeUtil.SERIF).deriveFont(letterStyle));
          } catch (NoFontFoundException e) {
            Font f;
            if(e.getFontList().length > 0)
              f = e.getFontList()[e.getBestFont()];
            else
              f = new Font("serif",Font.PLAIN,1);
            addSection(UnicodeUtil.fixString(toAdd,f),
                       f.deriveFont(letterStyle));
          }
          
          last = m.end();
        }

        //a non-letter section
        if(last < txt.length())
          try {
            toAdd = txt.substring(last,txt.length());
            addSection(toAdd,
                       UnicodeUtil.getDisplayFont
                       (toAdd,
                        UnicodeUtil.SERIF).deriveFont(Font.PLAIN));
          } catch (NoFontFoundException e) {
            Font f;
            if(e.getFontList().length > 0)
              f = e.getFontList()[e.getBestFont()];
            else
              f = new Font("serif",Font.PLAIN,1);
            addSection(UnicodeUtil.fixString(toAdd,f),
                       f.deriveFont(Font.PLAIN));
          }
      }
      
      public void addSection(String s, Font f) {
        if(head == null)
          head = tail = new TextSection(s,f);
        else
          tail = tail.next = new TextSection(s,f);
      }
      
      protected void setGraphics(Graphics g) {
        super.setGraphics(g);
        
        TextSection current = head;
        while(current != null) {
          current.createExtendedGlyphVector();
          current = current.next;
        }
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        TextSection current = head;
        while(current != null) {        
          current.draw(g,x,y);
          x += current.extendedGlyph.getWidth();
          current = current.next;
        }

        g.dispose();
      }
      
      protected double getWidth() {
        super.getWidth();

        double w = 0;
        
        TextSection current = head;
        while(current != null) {
          w += current.extendedGlyph.getWidth();
          current = current.next;
        }

        return w;
      }
      
      protected double getHeight() {
        super.getHeight();

        double h = 0;

        TextSection current = head;
        while(current != null) {
          if(current.extendedGlyph.getGlyphHeight() > h)
            h = current.extendedGlyph.getGlyphHeight();
          current = current.next;
        }

        return h;
      }

      public void setSize(float s) {
        int i = 0;
        try {
          TextSection current = head;
          while(current != null) {
            current.font = current.font.deriveFont(s);
            current = current.next;
          }
        } catch (Exception e) {
          System.out.println(e);
          System.out.println(tx);
          System.out.println("i = " + i);
          System.out.println("length = " + f.length);
          for(int j = 0; j < text.length; j++)
            System.out.print(text[j]);
          System.out.println();
        }
      }

      public float getSize() {
        return head.font.getSize();
      }
    }

    if(text.equals("="))
      return getEqualsSymbol();
    else
      return new TextSymbol(text, letterStyle);
  }

  /**
   * Returns a MathSymbol which is a fraction.
   *
   * @param n the numerator
   * @param d the denominator 
   * @return a MathSymbol which is a fraction
   */
  public static MathSymbol getFractionSymbol(String n, String d) {
    return getFractionSymbol(getTextSymbol(n).toMathExpression(),
                             getTextSymbol(d).toMathExpression());
  }

  /**
   * Returns a MathSymbol which is a fraction.
   *
   * @param n the numerator
   * @param d the denominator 
   * @return a MathSymbol which is a fraction
   */
  public static MathSymbol getFractionSymbol(MathExpression n,
                                             MathExpression d) {
    class FractionSymbol extends MathSymbol {
      MathExpression n;
      MathExpression d;

      //Font f;

      float s;

      float vgap;

      public FractionSymbol(MathExpression num, MathExpression den) {
        super();

        n = num;
        d = den;

        s = n.getSize();
      }

      protected void setGraphics(Graphics g) {
        super.setGraphics(g);

        n.setGraphics(g);
        d.setGraphics(g);

        vgap = (float)((n.getHeight() + d.getHeight())/4.0);
      }

      public void draw(float x, float y) {
        Graphics2D g = getDrawGraphics();

        n.draw(g,
               (float)(x + getWidth()/2f - n.getWidth()/2f),
               (float)(y - (d.getHeight() + vgap)));
        d.draw(g,
               (float)(x + getWidth()/2f - d.getWidth()/2f),
               y);

        float ly = (float)(y - (d.getHeight() + (vgap/2.0)));

        int strokeWidth = (int)Math.ceil((double)s/30.0);
        g.setStroke(new BasicStroke(strokeWidth));

        g.draw(new Line2D.Float
                 (x, ly,
                  (float)(x+getWidth()), ly));

        g.dispose();
      }
      
      public double getHeight() {
        super.getHeight();

        return n.getHeight() + d.getHeight() + vgap;
      }

      public double getWidth() {
        super.getWidth();

        return (n.getWidth() > d.getWidth()) ?
          n.getWidth() : d.getWidth();
      }

      public float getSize() {
        return s;
      }

      public void setSize(float f) {
        s = f;

        n.setSize(f*.6f);
        d.setSize(f*.6f);
      }
      
      public void setColor(Color c) {
        super.setColor(c);

        n.setColor(c);
        d.setColor(c);
      }

    }

    return new FractionSymbol(n,d);
  }

  /**
   * set the size of the MathSymbol
   *
   * @param s a point size value
   * @see #getSize
   */
  public abstract void setSize(float s);


  /**
   * Returns the point size of this MathSymbol.
   *
   * @return the point size of this MathSymbol
   * @see #setSize(float)
   */
  public abstract float getSize();

  /**
   * Returns a Font whose size is some percentage of the font given.
   *
   * @param font the font to scale
   * @param percent the percentage to scale the font by
   * @return a Font whose size is some percentage of font
   */
  protected Font scaleFont(Font font, double percent) {
    return font.deriveFont((float)(font.getSize() * (percent/100)));
  }

  /**
   * Returns the height of the passed GlyphVector according to visual
   * bounds.
   *
   * @param g the GlyphVector
   * @return the height of g
   */
  protected double getGlyphHeight(GlyphVector g) {
    if(System.getProperty("os.name").startsWith("Mac OS X"))
      return g.getOutline(0,0).getBounds().getHeight();
    else
      return g.getPixelBounds(null,0,0).getHeight();
  }

  /**
   * Returns the width of the passed GlyphVector according to visual
   * bounds.
   *
   * @param g the GlyphVector
   * @return the width of g
   */
  protected double getGlyphWidth(GlyphVector g) {
    if(System.getProperty("os.name").startsWith("Mac OS X"))
      return g.getOutline(0,0).getBounds().getWidth();
    else
      return g.getPixelBounds(null,0,0).getWidth();
  }

  /**
   * Returns the pixel width of a String according to the metrics of
   * the passed font.
   *
   * @param s a <code>String</code> value
   * @param f a <code>Font</code> value
   * @return the pixel width of the String
   */
  protected double getStringWidth(String s, Font f) {
    if(g2d == null)
      throw new IllegalStateException
        ("getStringWidth cannot be called when the Graphics object is not set.");      
    return g2d.getFontMetrics(f).getStringBounds(s,g2d).getWidth();
  }

  /**
   * Returns a MathExpression which contains only this symbol.
   *
   * @return a MathExpression which contains only this symbol
   */
  public MathExpression toMathExpression() {
    return new MathExpression(this,getSize());
  }

  /**
   * Returns a MathExpression which contains only this symbol. The
   * MathExpression will be based on the passed size.
   *
   * @param size the size of the MathExpression to return
   * @return a MathExpression which contains only this symbol
   */
  public MathExpression toMathExpression(float size) {
    return new MathExpression(this,size);
  }

  /**
   * Set the color of this symbol.
   *
   * @param c a <code>Color</code> value
   * @see #getColor()
   */
  public void setColor(Color c) {
    color = c;
  }

  /**
   * Returns the color of this symbol.
   *
   * @return the color of this symbol
   * @see #setColor(Color)
   */
  public Color getColor() {
    return color;
  }

  /**
   * Returns the Graphics2D object that should be used to draw this
   * <code>MathSymbol</code>. It is safe to transform or change this
   * graphics object without reverting it to its original state. Be
   * sure to dispose of this object before the 
   * {@link #draw(float,float)} method returns
   * (e.g. <code>g.dispose()</code>).
   *
   * @return the Graphics2D object that should be used to draw this
   * <code>MathSymbol</code>
   * @throws IllegalStateException if {@link #setGraphics(Graphics)}
   * has not yet been called.
   */
  protected Graphics2D getDrawGraphics() {
    if(g2d == null)
      throw new IllegalStateException
        ("set graphics has not been called yet");
        
    Graphics2D g = (Graphics2D) g2d.create();
    
    g.setColor(color);

    return g;
  }
}