package vcp.platform;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.net.URL;
import java.net.MalformedURLException ;

import java.util.Properties;
import java.util.HashMap;

import java.awt.Frame;

import javax.swing.JOptionPane;


import vcp.utilities.ExceptionDialog;

import vcp.VCP;

/**
 * This is a utility class with static methods that various parts of
 * the GUI need to use.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class GUIActions {

  /** the currently executing vcp object */
  public static VCP vcp;

  /** the configuration properties */
  protected static Properties config;

  /** not used? */
  protected static HashMap categoryPDFs;

  /*
   * load config properties
   */
  static {
    try {
      InputStream is = 
        GUIActions.class.getResourceAsStream("/vcp.cfg");

      config = new Properties();
      config.setProperty("pdf.viewer","default");
      config.setProperty("categories","Binomial,Fibonacci");
      config.load(is);

      String[] c = getCategories();
      
      categoryPDFs = new HashMap(c.length);

      for(int i = 0; i < c.length; i++) {
        categoryPDFs.put(c[i].toLowerCase(),config.getProperty(c[i] + ".pdf"));
      }
        
    } catch (Exception e) {
      ExceptionDialog.showExceptionDialog
        ("VCP: Configuration Error",
         "There was a problem loading the configuration file.\n" +
         "A default configuration will be used.",
	 e);
    }
    
    new Help();
    new About();    
  }
  
  /**
   * Display the html file at <code>url</code>. The <code>url</code>
   * can be either an actual URL or a path.
   *
   * @param url URL or path to HTML file
   */
  public static void displayHTML(String url) {
    try {
      if(vcp.isApplet()) {
	vcp.getAppletContext().showDocument
          (new URL(url),"_blank");
      } else {
        if(System.getProperty("os.name").equals("Linux")) {
          //under linux, the user will have to choose an
          //application to open pdf urls, for this demo, we will
          //use Adobe's reader
          Runtime.getRuntime().exec("firefox " + url);
        }
        
        else if(System.getProperty("os.name").startsWith("Windows")) {
          //under windows, we can pass the url name to cmd as a
          //command, and the url will be opened in the preferred
          //application
          Runtime.getRuntime().exec("cmd /C " + url);
        }
        
        else if(System.getProperty("os.name").startsWith("Mac OS X")) {
          //under Mac OS X, we use some applescript and the
          //osascript CLI application to get the Finder to open
          //the pdf in the preferred application
          Runtime.getRuntime().exec
            (new String[]
              {"osascript", "-e",
               "tell application \"Finder\"\r" +
               "open (POSIX url \"" + url + "\" as alias)\r" +
               "end tell"});
        }
        else
          JOptionPane.showMessageDialog(vcp.getFrame(),
                                        "No program is registered to " +
                                        "open HTML urls on " + 
                                        System.getProperty("os.name"), 
                                        "No Program Registered",
                                        JOptionPane.ERROR_MESSAGE);
      }
    } catch (Exception e) {
      ExceptionDialog.showExceptionDialog
        ("Cannot Open PDF","The PDF url, \"" + 
         url +"\" cannot be opened.",
	 e);
    }    
  }

  /**
   * Display the PDF file at the given path using the user's preferred
   * application.
   *
   * @param file the PDF file to display
   */
  public static void displayPDF(String file) {
    try {
      if(vcp.isApplet()) {
	vcp.getAppletContext().showDocument
          (new URL(vcp.getCodeBase(),file),"_blank");
      } else {
        file = System.getProperty("user.dir") + "/" + file;
        if((new File(file)).exists()) {
          if(config.getProperty("pdf.viewer").equals("default")) {
            if(System.getProperty("os.name").equals("Linux")) {
              //under linux, the user will have to choose an
              //application to open pdf files, for this demo, we will
              //use Adobe's reader
              Runtime.getRuntime().exec("xpdf " + file);
            }
            
            else if(System.getProperty("os.name").startsWith("Windows")) {
              //under windows, we can pass the file name to cmd as a
              //command, and the file will be opened in the preferred
              //application
              Runtime.getRuntime().exec("cmd /C " + "\""+file+"\"");
            }
            
            else if(System.getProperty("os.name").startsWith("Mac OS X")) {
              //under Mac OS X, we use some applescript and the
              //osascript CLI application to get the Finder to open
              //the pdf in the preferred application
              Runtime.getRuntime().exec
                (new String[]
                  {"osascript", "-e",
                   "tell application \"Finder\"\r" +
                   "open (POSIX file \"" + file + "\" as alias)\r" +
                   "end tell"});
            }
            else
              JOptionPane.showMessageDialog(vcp.getFrame(),
                                            "No program is registered to " +
                                            "open PDF files on " + 
                                            System.getProperty("os.name"), 
                                            "No Program Registered",
                                            JOptionPane.ERROR_MESSAGE);
          } else {
            String[] cmd;

            if(System.getProperty("os.name").startsWith("Mac OS X")) {
              cmd = new String[] 
                {
                  "open", 
                  "-a", 
                  config.getProperty("pdf.viewer"),
                  file
                };
            }
            
            else {
              cmd =new String[] 
                { 
                  config.getProperty("pdf.viewer"),
                  file
                };
            }
            
            Runtime.getRuntime().exec(cmd);
          }
	
        } else {
          JOptionPane.showMessageDialog(vcp.getFrame(),file + " does not exist",
                                        "PDF file does not exist", 
                                        JOptionPane.ERROR_MESSAGE);
        }
      }
    } catch (Exception e) {
      ExceptionDialog.showExceptionDialog
	("Cannot Open PDF","The PDF file, \"" + 
	 file +"\" cannot be opened.",
	 e);
    }
  }

  /**
   * Load a module
   */
  public static void loadModule() {
    Module current;

    if((current = ModuleChooser.showModuleChooser(vcp.getFrame())) != null) {
      //System.out.println("pdf");
      displayPDF("modules/" + current.getLocation() + 
                 current.getProperty("exposition"));

      //System.out.println("panel");
      ModulePanel modPanel = new ModulePanel(current);
      //System.out.println("set panel");
      vcp.setMainPanel(modPanel);
      //System.out.println("title");
      
      vcp.setTitle("VCP - " + current.getProperty("name"));
    }
  }

  /**
   * pack the main window
   */
  public static void pack() {
    vcp.pack();
  }

  /**
   * Initialize the GUIActions class (should only be called
   * <em>once</em> per run)
   *
   * @param v a <code>VCP</code> value
   */
  public static void initialize(VCP v) {
    vcp = v;

    if(!vcp.isApplet()) {
      try {
        config.load(new FileInputStream("vcp.cfg"));
      } catch (FileNotFoundException fnfe) {
        //do nothing
      } catch (Exception e) {
        ExceptionDialog.showExceptionDialog
          ("VCP: Configuration Error",
           "There was a problem loading the configuration file.\n" +
           "A default configuration will be used.",
           e);
      }
    }
  }

  /**
   * Get the root of the application (either the user directory or the
   * codebase of the applet)
   *
   * @return the root of the application
   * @see #getRootAsString()
   */
  public static URL getRoot() {
    if(vcp.isApplet()) {
      return vcp.getCodeBase();
    } else {
      try {
        return new URL("file:" + System.getProperty("user.dir") + "/");
      } catch (MalformedURLException e) { 
        return null;
      }
    }
  }

  /**
   * Get the root of the application (either the user directory or the
   * codebase of the applet) as a string. If we are an applet, the
   * string is a url; if we are an application, the string is a path.
   *
   * @return the root of the application
   * @see #getRoot()
   */
  public static String getRootAsString() {
    if(vcp.isApplet())
      return vcp.getCodeBase().toString();
    else
      return System.getProperty("user.dir") + 
        System.getProperty("file.separator");
  }

  /**
   * show the basic definitions file
   */
  public static void showBasicDefs() {
    displayPDF("modules/definitions/CommonDefs.pdf");
  }


  /**
   * Returns the module categories
   *
   * @return the module categories
   */
  public static String[] getCategories() {
    return config.getProperty("categories").split(",");
  }

  /*
  public static String getCategoryPDF(String p) {
    return (String) categoryPDFs.get(p.toLowerCase());
  }
  */

  /**
   * Return the value of the configuration property
   *
   * @param p the property whose value is desired
   * @return the value of the configuration property
   */
  public static String getProperty(String p) {
    return config.getProperty(p);
  }

  /**
   * Set a configuration properties value
   *
   * @param p the property to set
   * @param v the value to set it to
   */
  public static void setProperty(String p, String v) {
    config.setProperty(p,v);
  }
    
}
