package vcp.platform;

import javax.swing.JPanel;
import javax.swing.JDialog;
import javax.swing.JList;
import javax.swing.JTree;
import javax.swing.JTextArea;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;

import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.event.TreeSelectionListener;

import javax.swing.event.TreeSelectionEvent;

import javax.swing.border.LineBorder;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import java.util.Hashtable;

import vcp.utilities.ExceptionDialog;
import vcp.utilities.ScrollableTextArea;

import info.clearthought.layout.TableLayout;

import java.io.*;
import java.net.*;


/**
 * A GUI frame used to load modules
 *
 * @author phil kaasa
 * @version 1.0
 */
public class ModuleChooser extends JDialog implements 
                                            TreeSelectionListener,
                                            ActionListener {
  /** value if OK clicked */
  public static int OK = 0;

  /** value if Cancel clicked */
  public static int CANCEL = 1;

  /** the thread that loads the modules */
  static Thread moduleLoadThread;
  
  /** the root of the tree displaying modules */
  static DefaultMutableTreeNode moduleRoot;

  /** the tree to display the available modules */
  JTree moduleTree;

  /** text area to display module summaries */
  ScrollableTextArea summary;

  /** the name of the currently selected module */
  JLabel name;

  /** the authors of the currently selected module */
  JLabel authors;

  /** the lable to show that modules are currently being loaded */
  JLabel loadingLabel;
  
  /** cancel button */
  JButton cancelButton;

  /** ok button */
  JButton okButton;

  /** the value of the chooser */
  int retVal = CANCEL;

  /** the currently selected module */
  Module current;

  /**
   * Creates a new <code>ModuleChooser</code> instance.
   * @param f the parent <code>Frame</code> (can be
   * <code>null</code>).
   */
  protected ModuleChooser(Frame f) {
    super(f, "Choose a Module");

    long start = System.currentTimeMillis();

    current = null;

    setModal(true);

    final JPanel content = new JPanel();
    setContentPane(content);

    TableLayout tl = new TableLayout
      (new double[][] {
        {TableLayout.PREFERRED,
         TableLayout.FILL,
         TableLayout.PREFERRED, 
         TableLayout.PREFERRED},
        {TableLayout.PREFERRED,
         TableLayout.PREFERRED,
         TableLayout.FILL,
         TableLayout.PREFERRED}
      });

    tl.setHGap(8);
    tl.setVGap(8);

    content.setLayout(tl);

    content.setBorder(new LineBorder(content.getBackground(),8));

    summary = new ScrollableTextArea();
    summary.getJScrollPane().setPreferredSize(new Dimension(400,200));
    
    name = new JLabel();
    name.setFont(name.getFont().deriveFont(16f));
    authors = new JLabel();
    authors.setFont(authors.getFont().deriveFont(10f));
    
    cancelButton = new JButton("Cancel");
    okButton = new JButton("OK");

    cancelButton.addActionListener(this);
    okButton.addActionListener(this);

    //SwingUtilities.invokeLater(new Runnable() {
    //    public void run() {
          content.add(name, "1,0,3,0");
          content.add(authors, "1,1,3,1");
          
          content.add(loadingLabel = new JLabel("Loading\nModules..."),
                      "0,0,0,2");
          
          content.add(summary.getJScrollPane(),"1,2,3,2");
          content.add(cancelButton,"2,3");
          content.add(okButton,"3,3");

          setDefaultCloseOperation(DISPOSE_ON_CLOSE);
          pack();
          //}
          //});

    //(new Thread(new Runnable() {
    SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          try {
            moduleLoadThread.join();
          } catch (InterruptedException e) {}
          
          moduleTree = new JTree(moduleRoot);
          DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
          moduleTree.setRootVisible(false);
          ((javax.swing.tree.DefaultTreeModel)
           moduleTree.getModel()).setAsksAllowsChildren(true);
          moduleTree.getSelectionModel().setSelectionMode
            (javax.swing.tree.TreeSelectionModel.SINGLE_TREE_SELECTION);
          for(int i = 0; i < moduleTree.getRowCount(); i++)
            moduleTree.expandRow(i);
          moduleTree.addTreeSelectionListener(ModuleChooser.this);
          
          getContentPane().remove(loadingLabel);
          getContentPane().add
            (new JScrollPane(moduleTree,
                             JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                             JScrollPane.HORIZONTAL_SCROLLBAR_NEVER),
             "0,0,0,2");
          
          validate();
          repaint();
          
          //fix resizing annoyance
          moduleTree.setMinimumSize(moduleTree.getSize());
          moduleTree.setPreferredSize(moduleTree.getSize());            
        }
      });
        //})).start();
  }

  /**
   * Called when the selected module changes. Loads that modules
   * information in the summary section.
   *
   * @param e a <code>TreeSelectionEvent</code> value
   */
  public void valueChanged(TreeSelectionEvent e) {
    DefaultMutableTreeNode n = (DefaultMutableTreeNode) 
      e.getPath().getLastPathComponent();

    if(n.getUserObject() instanceof Module) {
      current = (Module) n.getUserObject();      
      name.setText(current.getProperty("name"));
      authors.setText(current.getProperty("author.exposition.name") +
                      ", " +
                      current.getProperty("author.visualization.name"));
      summary.setContent(current.getProperty("abstract"));
    }
  }

  /**
   * Handels button and other actions.
   *
   * @param e an <code>ActionEvent</code> value
   */
  public void actionPerformed(ActionEvent e) {
    if(e.getSource() == okButton) {
      retVal = OK;
      hideModuleChooser();
    } else if(e.getSource() == cancelButton) {
      retVal = CANCEL;
      hideModuleChooser();
    }
  }

  /**
   * Initialize the module data.
   */
  public synchronized static void initialize() {
    if(moduleRoot == null) {
      moduleLoadThread = new Thread(new Runnable() {
          public void run() {
            loadModules();
          }
        });
      
      moduleLoadThread.start();
    }    
  }

  /**
   * Display a module chooser. If this is being called for the first
   * time, the available modules will first be loaded. Returns the
   * selected module if OK is pressed, null otherwise.
   *
   * @param parent a <code>Frame</code> value (can be
   * <code>null</code>)
   * @return the selected module if OK is pressed, null otherwise
   */
  public static Module showModuleChooser(Frame parent) {
    initialize();
    
    final ModuleChooser moduleChooser = new ModuleChooser(parent);
    moduleChooser.setValue(CANCEL);

    moduleChooser.setVisible(true);
    if(moduleChooser.getValue() == OK)
      return moduleChooser.current;
    else
      return null;
  }

  /**
   * set the value of the ModuleChooser
   *
   * @param val an <code>int</code> value
   */
  protected void setValue(int val) {
    retVal = val;
  }

  /**
   * Returns the value of the ModuleChooser
   *
   * @return the value of the ModuleChooser
   */
  protected int getValue() {
    return retVal;
  }

  /**
   * Hide the module chooser
   */
  protected void hideModuleChooser() {
    //setVisible(false);
    dispose();
  }

  /**
   * Discover available modules and load them.
   */
  protected static void loadModules() {
    moduleRoot = new DefaultMutableTreeNode();
      
    String[] categoryStrings = GUIActions.getCategories();
    
    Hashtable categories = new Hashtable(categoryStrings.length);
    
    
    for(int i = 0; i < categoryStrings.length; i++) {
      DefaultMutableTreeNode temp;
      categories.put(categoryStrings[i].toLowerCase(),temp =
                     new DefaultMutableTreeNode(categoryStrings[i]));
      temp.setAllowsChildren(true);
      moduleRoot.add(temp);
    }

    Module moduleT;

    try {
      String dir;
      BufferedReader br;

      if(GUIActions.vcp.isApplet())
        br = new BufferedReader
          (new InputStreamReader
           (ModuleChooser.class.getResource
            ("/modules/module_list").openStream()));
      else
        br = new BufferedReader
          (new InputStreamReader
           ((new URL(GUIActions.getRoot(),"modules/")).openStream()));
      
      while((dir = br.readLine()) != null) {
        try {
          InputStream is;

          if(GUIActions.vcp.isApplet())
            is = ModuleChooser.class.getResource
              ("/modules/" + dir + "/module.cfg").openStream();
          else
            is = (new URL(GUIActions.getRoot(),"modules/" + dir + 
                          "/module.cfg")).openStream();
          moduleT = new Module(dir);
          moduleT.load(is);
          DefaultMutableTreeNode node = new DefaultMutableTreeNode(moduleT);
          node.setAllowsChildren(false);
          if(categories.get(moduleT.getProperty("category")) != null)
            ((DefaultMutableTreeNode) 
             categories.get(moduleT.getProperty("category").toLowerCase())).
              add(node);
          else
            moduleRoot.add(node);
        } catch (FileNotFoundException fnfe) {
          //ignore this error and move on to the next directory
        }
      }
      
    } catch (IOException ioe) {
      ExceptionDialog.showExceptionDialog
        ("Module Load Error",
         "There was a problem loading modules.\nThe program will now exit.",
         ioe);
      System.exit(1);
    }
  }

  /**
   * for testing
   *
   * @param args a <code>String[]</code> value
   */
  public static void main(String[] args) {
    new vcp.VCP();

    System.setProperty
      ("user.dir",
       System.getProperty("user.dir") + "/vcp");

    System.out.println(ModuleChooser.showModuleChooser(null));
    System.exit(0);
  }
}