package vcp.platform;

import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JSeparator;
import javax.swing.BoxLayout;
import javax.swing.Box;

import javax.swing.border.LineBorder;

import java.awt.Color;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.File;

import java.net.URLClassLoader;
import java.net.URL;

import info.clearthought.layout.TableLayout;

import vcp.utilities.ExceptionDialog;

/**
 * A GUI panel to display a module's visualization
 *
 * @author phil kaasa
 * @version 1.0
 */
public class ModulePanel extends JPanel implements ActionListener {
  /** button to display common definitions */
  JButton defs;

  //JButton categoryDefs;

  /** button to display exposistion */
  JButton exposition;

  /** button to restart visualization */
  JButton startOver;

  /** button to start visualization */
  JButton startVisualization;

  /** button to load another module */
  JButton loadAnotherModule;

  /** button to show about VCP pdf */
  JButton aboutVCP;

  /** button to show about combinatorial proofs pdf */
  JButton aboutComb;

  /** the module whose visualization is being displayed */
  Module module;

  /** the panel to contain the visualization */
  JPanel modulePanel;

  /** the class of the module's visualization */
  Class modulePanelClass;

  /** the thread loading the module */
  Thread moduleLoadThread;

  /**
   * Creates a new <code>ModulePanel</code> instance with no loaded
   * module. (This is used to create the panel when the VCP
   * application first executes.)
   */
  public ModulePanel() {
    super();
    
    modulePanel = null;

    layoutPanel();

    startVisualization.setEnabled(false);
    java.awt.Dimension d = startVisualization.getPreferredSize();
    startVisualization.setText("");
    startVisualization.setPreferredSize(d);
    startVisualization.setSize(d);
  }

  /**
   * Creates a new <code>ModulePanel</code> instance.
   *
   * @param m a <code>Module</code> value
   */
  public ModulePanel(Module m) {
    super();

    module = m;

    modulePanel = null;

    moduleLoadThread = new Thread(new Runnable() {
	public void run() {
	  loadModuleMainClass();
	}
        });

    moduleLoadThread.start();
    
    layoutPanel();

    exposition.setEnabled(true);
    defs.setEnabled(true);
  }

  /**
   * Create the GUI components and layout the panel.
   */
  private void layoutPanel() {
    TableLayout tl = new TableLayout
      (new double[][] {
        {TableLayout.PREFERRED,
         TableLayout.PREFERRED,
	 TableLayout.FILL},
        {TableLayout.PREFERRED,
         TableLayout.FILL}
      });
    
    tl.setHGap(8);
    tl.setVGap(8);
    
    setLayout(tl);
    
    defs = new JButton("Common Definitions");
    //categoryDefs = new JButton("Category Definitions");
    exposition = new JButton("Exposition");
    startOver = new JButton("Restart Visualization");
    loadAnotherModule = new JButton("Load Module");
    startVisualization = new JButton("<html>The exposition of the proof should open in another window.<br><br>Click here to display the visualization when you are ready.</html>");
    aboutVCP = new JButton("About VCP");
    aboutComb = new JButton("About Combinatorial Proofs");
    
    defs.addActionListener(this);
    //categoryDefs.addActionListener(this);
    exposition.addActionListener(this);
    startOver.addActionListener(this);
    loadAnotherModule.addActionListener(this);
    startVisualization.addActionListener(this);
    aboutVCP.addActionListener(this);
    aboutComb.addActionListener(this);
    
    startOver.setEnabled(false);
    exposition.setEnabled(false);
    defs.setEnabled(false);
    
    JPanel buttonPanel = new JPanel();
    //buttonPanel.setBorder(new LineBorder(Color.black,2));
    buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.Y_AXIS));

    java.awt.Dimension max = 
      new java.awt.Dimension(Integer.MAX_VALUE,Integer.MAX_VALUE);

    defs.setMaximumSize(max);
    //categoryDefs.setMaximumSize(max);
    exposition.setMaximumSize(max);
    startOver.setMaximumSize(max);
    loadAnotherModule.setMaximumSize(max);
    aboutVCP.setMaximumSize(max);
    aboutComb.setMaximumSize(max);

    java.awt.Dimension space = new java.awt.Dimension(0,20);

    buttonPanel.add(defs);
    buttonPanel.add(exposition);
    buttonPanel.add(startOver);
    buttonPanel.add(Box.createRigidArea(space));
    buttonPanel.add(loadAnotherModule);
    buttonPanel.add(Box.createRigidArea(space));
    buttonPanel.add(aboutVCP);
    buttonPanel.add(aboutComb);

    add(startVisualization,"1,0,2,1");
    add(buttonPanel, "0,0");    
  }


  /**
   * Load the module's visualization class and add it to the panel
   */
  public void loadModuleMainClass() {
    try {
      //get the appropriate module class loader depending on the system
      ClassLoader l;
      if(GUIActions.vcp.isApplet()) {
        modulePanelClass = getClass().forName//l.loadClass
          (module.getProperty("main.class"));
      } else {
        l = new ModuleClassLoader();
        modulePanelClass = l.loadClass
          (module.getProperty("main.class"));
      }
      
      modulePanel = (JPanel) modulePanelClass.newInstance();
    } catch (Exception e) {
      ExceptionDialog.showExceptionDialog
	("Module Error",
	 "An error ocurred while loading the module.",
	 e);
    }
  }

  /**
   * Display the module's visualization panel
   */
  private void displayModulePanel() {
    //wait until the modulePanel has been loaded
    try {
      if (modulePanel == null) {
        startVisualization.setText("Loading Module...");
        moduleLoadThread.join();
      }

      remove(startVisualization);
      add(modulePanel,"1,0,2,1");
      validate();
      GUIActions.pack();
      modulePanel.validate();
      modulePanel.repaint();
      
      startOver.setEnabled(true);
    } catch (InterruptedException ie) {
      ExceptionDialog.showExceptionDialog
        ("Module Error",
         "There was a problem displaying the visualization, try clicking the button again.",
         ie);
    } catch (Exception e) {
      ExceptionDialog.showExceptionDialog
        ("Error",
         "There was a problem displaying the visualization.",
         e);      
    }
  }

  /**
   * handles button actions.
   *
   * @param e an <code>ActionEvent</code> value
   */
  public void actionPerformed(ActionEvent e) {
    Object ob = e.getSource();

    if(ob == startVisualization) {
      displayModulePanel();
    }

    else if (ob == exposition) {
      GUIActions.displayPDF("modules/" + module.getLocation() + 
                            module.getProperty("exposition"));
    }

    else if (ob == defs) {
      GUIActions.showBasicDefs();
    }

    else if (ob == startOver){
      try {
	remove(modulePanel);
	modulePanel = (JPanel) modulePanelClass.newInstance();
	add(modulePanel,"1,0,2,1");
	validate();
	GUIActions.pack();
	modulePanel.validate();
	modulePanel.repaint();        
      } catch (Exception ex) {
	ExceptionDialog.showExceptionDialog
          ("Module Load Error",
           "There was an error reloading the module",
           ex);
      }
    }

    else if (ob == loadAnotherModule) {
      GUIActions.loadModule();
    }

    else if (ob == aboutVCP) {
      GUIActions.displayPDF
        ("modules/definitions/ProjectIntro.pdf");
    }

    else if (ob == aboutComb) {
      GUIActions.displayPDF
	("modules/definitions/CombProofIntro.pdf");      
    }
  } 
}