package vcp.utilities;

import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import java.awt.Frame;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

/**
 * A class for a dialog to display when exceptions are thrown.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class ExceptionDialog extends JDialog implements PropertyChangeListener {
  /** The JOptionPane for the dialog */
  private JOptionPane optionPane;

  /** The Throwable that caused this dialog to be displayed */
  private Throwable excp;

  /** The JScrollPane for the JTextArea which contains the details of
   * the exception. */
  private JScrollPane exceptionDetails;


  /**
   * Creates a new <code>ExceptionDialog</code> instance. This method
   * can only be called by showExceptionDialog().
   *
   * @param title The title of the dialog
   * @param msg The message that the user sees
   * @param e The exception that caused the dialog to be displayed
   * @param parent The Frame from which the dialog is displayed
   */
  private ExceptionDialog(String title, String msg, String devMsg, 
                          Throwable e, Frame parent) {
    super(parent,title,true);
    createDialog(title, msg, devMsg, e);
  }
  
  /**
   * Creates a new <code>ExceptionDialog</code> instance. This method
   * can only be called by showExceptionDialog().
   *
   * @param title The title of the dialog
   * @param msg The message that the user sees
   * @param e The exception that caused the dialog to be displayed
   */
  private ExceptionDialog(String title, String msg, String devMsg, 
                          Throwable e) {
    super((Frame) null,title,true);
    createDialog(title, msg, devMsg, e);
  }
  
  /**
   * This method is called from all the constructers.
   *
   * @param title the title of the dialog
   * @param msg the message to display to the user
   * @param e The exception that caused the dialog to be displayed
   */
  private void createDialog(String title, String msg, String devMsg, 
                            Throwable e) {
    excp = e;

    StringWriter sw = new StringWriter();
    excp.printStackTrace(new PrintWriter(sw));

    JTextArea ta = new JTextArea
      (((devMsg == null) ? "" : (devMsg + "\n\n")) + 
       sw.toString(),6,45);
    ta.setEditable(false);

    exceptionDetails = new JScrollPane
      (ta,
       JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
       JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

    
    exceptionDetails.setVisible(false);

    Object[] array = {msg, exceptionDetails};

    Object[] options = {"OK","Details"};

    optionPane = new JOptionPane(array,
				 JOptionPane.ERROR_MESSAGE,
				 JOptionPane.OK_CANCEL_OPTION,
				 null,
				 options,
				 options[0]);

    optionPane.addPropertyChangeListener(this);
    setContentPane(optionPane);

    setDefaultCloseOperation(DISPOSE_ON_CLOSE);
    setResizable(false);
    
    pack();

    setVisible(true);
  }

  /**
   * Required for PropertyChangeListener. This is called when the
   * value of the optionPane changes. It handles what occurs when the
   * OK or Details buttons are clicked.
   *
   * @param e a <code>PropertyChangeEvent</code> value
   */
  public void propertyChange(PropertyChangeEvent e) {
    if(isVisible()) {
      
      Object value = optionPane.getValue();
      optionPane.setValue(JOptionPane.UNINITIALIZED_VALUE);

      if(value.equals("OK")) {
	setVisible(false);
	dispose();
      } else if(value.equals("Details")) {
	exceptionDetails.setVisible(!exceptionDetails.isVisible());
        exceptionDetails.revalidate();
	pack();
      }
    }

  }  

  /**
   * Creates and displays a new <code>ExceptionDialog</code>
   * instance.
   *
   * @param title The title of the dialog
   * @param msg The message that the user sees
   * @param e The exception that caused the dialog to be displayed
   * @param parent The Frame from which the dialog is displayed
   */
  public static void showExceptionDialog(String title, String msg, 
                                         String devMsg,
                                         Throwable e, Frame parent) {
    new ExceptionDialog(title, msg, devMsg, e, parent);
  }

  /**
   * Creates and displays a new <code>ExceptionDialog</code>
   * instance.
   *
   * @param title The title of the dialog
   * @param msg The message that the user sees
   * @param devMsg An extra message to be displayed in the details
   * before the stacktrace
   * @param e The exception that caused the dialog to be displayed
   */
  public static void showExceptionDialog(String title, String msg, 
                                         String devMsg,
                                         Throwable e) {
    new ExceptionDialog(title, msg, devMsg, e);
  }

  /**
   * Creates and displays a new <code>ExceptionDialog</code>
   * instance.
   *
   * @param title The title of the dialog
   * @param msg The message that the user sees
   * @param e The exception that caused the dialog to be displayed
   */
  public static void showExceptionDialog(String title, String msg, 
                                         Throwable e) {
    new ExceptionDialog(title, msg, null, e);
  }  

  /**
   * Creates and displays a new <code>ExceptionDialog</code>
   * instance.
   *
   * @param title The title of the dialog
   * @param msg The message that the user sees
   * @param e The exception that caused the dialog to be displayed
   * @param parent The Frame from which the dialog is displayed
   */
  public static void showExceptionDialog(String title, String msg, 
                                         Throwable e, Frame parent) {
    new ExceptionDialog(title, msg, null, e, parent);
  }

  /**
   * A main method for testing.
   *
   * @param args a <code>String[]</code> value
   */
  public static void main(String[] args) {
    ExceptionDialog.showExceptionDialog
      ("test", "error", new Exception("blah, blah, blah"), null);
  }
}