package vcp.utilities;

import javax.swing.JPanel;
import javax.swing.JComponent;
import javax.swing.JTextArea;
import javax.swing.JScrollPane;
import javax.swing.JScrollBar;
import javax.swing.SwingUtilities;
import javax.swing.BoundedRangeModel;

import javax.swing.border.LineBorder;
import javax.swing.text.Utilities;

import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Dimension;
import java.awt.Color;

import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.*;

import java.util.Vector;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import vcp.platform.ModuleClassLoader;

import vcp.mathnotation.MathExpression;
import vcp.mathnotation.MathSymbol;

import info.clearthought.layout.TableLayout;


/**
 * ScrollableTextArea is a JPanel which can contain interleaved text
 * and MathExpressions. It has methods to allow programmatic scrolling
 * to points between content blocks.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class ScrollableTextArea extends JPanel 
  implements ComponentListener, ScrollableTextAreaConstants,
             AdjustmentListener {

  /** a Vector containing the contents of the contentPanel */
  protected Vector contents;

  /** the layout of the contents */
  protected TableLayout tl;

  /** the mark to scroll to if {@link #scrollToMark()} is called */
  protected int scrollMark;

  /** the JScrollPane for this ScrollableTextArea */
  protected JScrollPane sp;

  protected int debugNum = 0;

  /**
   * a subclass of JPanel which displays a single MathExpression
   */
  protected class ExpressionPanel extends JPanel {
    MathExpression expr;
    int margin;

    public ExpressionPanel(MathExpression e) {
      expr = e;

      margin =(int) (ScrollableTextArea.this.getFontMetrics
                     (ScrollableTextArea.this.getFont()).getHeight() / 2.0);

      setBackground(Color.WHITE);
    }

    public void addNotify() {
      super.addNotify();

      Graphics g = getGraphics();

      setPreferredSize(new Dimension((int)expr.getWidth(g),
                                     (int)expr.getHeight(g)+margin*2));
    }
      
    public void paintComponent(Graphics g) {
      super.paintComponent(g);

      float x = 
        (float)ScrollableTextArea.this.getWidth()/2.0f - 
        (float)expr.getWidth(g)/2.0f;

      float y = (float) (getHeight()/2f + expr.getHeight(g)/2f);
      
      expr.draw(g,x,y);      
    }
  }

  /**
   * Creates a new <code>ScrollableTextArea</code> instance.
   *
   * @param vp vertical scrollbar policy
   * @param hp horizontal scrollbar policy
   */
  public ScrollableTextArea(int vp, int hp) {
    setLayout(tl = new TableLayout(new double[][]
      {{TableLayout.FILL},{TableLayout.FILL}}));

    try {
      setFont(UnicodeUtil.getAlphaNumberFont().
              deriveFont(14f));
    } catch (NoFontFoundException e) {
      setFont(new Font("Serif", Font.PLAIN, 14)); 
    }

    setBackground(Color.WHITE);

    sp = new JScrollPane(this,vp,hp);
    sp.getVerticalScrollBar().getModel().setMaximum(Integer.MAX_VALUE);
    sp.getViewport().addComponentListener(this);
    sp.getVerticalScrollBar().addAdjustmentListener(this);


    contents = new Vector(5,2);
  }

  /**
   * Creates a new <code>ScrollableTextArea</code> instance.
   */
  public ScrollableTextArea() {
    this(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
         JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
  }

  public void componentHidden(ComponentEvent e) {

  }

   public void componentMoved(ComponentEvent e) {

  }

  public void componentResized(ComponentEvent e) {
    //System.out.println(getSize());
    //System.out.print(sp.getViewport().getExtentSize().getWidth() + "  ");
    //SwingUtilities.invokeLater(new Runnable() {
    //    public void run() {
          Dimension d = new Dimension
            ((int)Math.floor(sp.getViewport().getExtentSize().getWidth()),
             getContentsHeight());
          
          setPreferredSize(d);
          setSize(d);
          //System.out.print(sp.getViewport().getExtentSize().getWidth() + "  ");
          //System.out.println(getSize().getWidth());
          invalidate();
          validate();
          //System.out.println("-               resize");
          //}
          //});
  }

  public void componentShown(ComponentEvent e) {

  }

  /**
   * Add text to this ScrollableTextArea. The text will not be
   * processed in any way.
   *
   * @param s the text to add
   */
  protected synchronized void add(final String s) {
    final int n = debugNum++;
    //System.out.println(n + ": start add s");

    //SwingUtilities.invokeLater(new Runnable() {
    //    public void run() {
    //System.out.println(n + ": adding s ");            
          final Object c;

          if((contents != null) && (contents.size() > 0) &&
             (c = contents.get(contents.size()-1)) instanceof JTextArea) {
            ((JTextArea) c).append(s);
            ((JTextArea) c).validate();
            //System.out.println("app s");
          } else {
            final JTextArea ta = new JTextArea(s);
            setTextAreaSettings(ta);
            
            tl.insertRow(contents.size(), TableLayout.PREFERRED);
            add(ta,"0," + contents.size());
            contents.add(ta);
            componentResized(null);
          }
          //System.out.println(n + ": added s ");            
          //}
          //});
    sp.invalidate();
    sp.validate();
    sp.repaint();
    //System.out.println(n+ ": return from add s ");
  }

  /**
   * Adds the passed <code>MathExpression</code> to the contents.
   *
   * @param e the MathExpression to add.
   */
  protected synchronized void add(final MathExpression e) {
    final int n = debugNum++;
    //System.out.println(n+ ": start add m ");
   
    //SwingUtilities.invokeLater(new Runnable() {
    //    public void run() {
    //System.out.println(n + ": adding m");            
          ExpressionPanel ep = new ExpressionPanel(e);
          
          tl.insertRow(contents.size(), TableLayout.PREFERRED);
          add(ep,"0," + contents.size());
          contents.add(ep);
          componentResized(null);

          //System.out.println(n+ ": added m ");
          
          sp.invalidate();
          sp.validate();
          sp.repaint();
          //    }
  //});
          //System.out.println(n + ": return from add m ");
  }

  /**
   * Add content to this ScrollableTextArea. The passed string will be
   * processed to add included MathExpressions if any.
   *
   * @param s the content to add
   */
  protected void addContent(String s, int actions) {
    //finds things that look like: \[expression.class\]. Will not
    //match "\\[" or "\\]"
    
    String expr = "(.*?[^\\\\])\\\\\\[(.*?[^\\\\])\\\\\\]";
    
    Matcher m = (Pattern.compile(expr)).matcher(s);

    int end = 0;
    while(m.find()) {
      if(m.group(1).length() > 0)
        //replace "\\[" or "\\]" with "\[" or "\]"
        add(m.group(1).replaceAll("\\\\\\\\([\\[\\]])","\\\\$1"));
      
      //do class loading to add mathexpression instead
      try {
        try {
          add((MathExpression)getClass().forName(m.group(2)).newInstance());
        } catch(ClassNotFoundException cnfe) {
          MathExpression mexpr = (MathExpression) 
            (new ModuleClassLoader()).loadClass
            (m.group(2)).newInstance();
          add(mexpr);
        }
      } catch (Exception e) {
        add("\\[" + m.group(2) + "\\]");
        ExceptionDialog.showExceptionDialog
          ("Mathematical Expression Error",
           "This module's summary has a mathematical expression in " +
           "it that cannot be displayed.",e);
      }

      end = m.end();
    }
    
    if(s.substring(end).length() > 0)
      add(s.substring(end).replaceAll("\\\\\\\\([\\[\\]])","\\\\$1"));
  }

  /**
   * Clear the contents and set the scroll bar to the top.
   */
  public void clear() {
    contents = new Vector(5,2);
    removeAll();
    setLayout(tl = new TableLayout(new double[][]
      {{TableLayout.FILL},{TableLayout.FILL}}));

    sp.getVerticalScrollBar().setValue(0);
  }

  /**
   * Set the text content of this ScrollableTextArea. This method
   * clears any content, then adds the content represented by the
   * passed <code>String</code>. The <code>String</code> will be
   * parsed for expressions in \[ \] tags.
   *
   * @param s the text to set
   * @param actions the only action this method accepts is
   * {@link ScrollableTextAreaConstants#NO_UPDATE}
   */
  public void setContent(String s, int actions) {
    clear();
    
    addContent(s, 0);

    if((actions & NO_UPDATE) == 0)
      update();
  }

  /**
   * Set the text content of this ScrollableTextArea. This method
   * clears any content, then adds the content represented by the
   * passed <code>String</code>. The <code>String</code> will be
   * parsed for expressions in \[ \] tags.
   *
   * @param s the text to set
   */
  public void setContent(String s) {
    setContent(s,0);
  }

  /**
   * Set the content of this ScrollableTextArea. This method clears
   * any content, then adds the passed <code>MathExpression</code>.
   *
   * @param expr the <code>MathExpression</code> to set as content
   * @param actions the only action this method accepts is
   * {@link ScrollableTextAreaConstants#NO_UPDATE}
   */
  public void setContent(MathExpression expr, int actions) {
    clear();

    add(expr);

    if((actions & NO_UPDATE) == 0)
      update();
  }

  /**
   * Set the content of this ScrollableTextArea. This method clears
   * any content, then adds the passed <code>MathExpression</code>.
   *
   * @param expr the <code>MathExpression</code> to set as content
   */
  public void setContent(MathExpression expr) {
    setContent(expr,0);
  }

  /**
   * Append the passed string to the content. The <code>String</code>
   * will be parsed for expressions in \[ \] tags.
   *
   * @param s the text to append 
   * @param actions action to while appending the text. One or more of 
   * {@link ScrollableTextAreaConstants#SET_MARK}, 
   * {@link ScrollableTextAreaConstants#NEWLINES}, 
   * {@link ScrollableTextAreaConstants#NO_UPDATE}, 
   * {@link ScrollableTextAreaConstants#SCROLL}.
   * These can be strung together with bitwise OR, "|".
   */
  public synchronized void append(String s, int actions) {
    //System.out.println("append s");

    boolean isTa = ((contents != null) && (contents.size() > 0) &&
                    (contents.get(contents.size()-1)) instanceof JTextArea);

    if((actions & NEWLINES) == NEWLINES && isTa)
      add("\n");

    if((actions & SET_MARK) == SET_MARK) {
      setScrollMark();
    }

    int m = sp.getVerticalScrollBar().getValue();

    if((actions & NEWLINES) == NEWLINES && isTa)
      add("\n");

    addContent(s, actions);
    //add(s);

    if((actions & SCROLL) == SCROLL)
      scrollToMark();
    else
      sp.getVerticalScrollBar().setValue(m);

    if((actions & NO_UPDATE) == 0)
      update();
  }

  /**
   * Append the passed string to the content. The <code>String</code>
   * will be parsed for expressions in \[ \] tags.
   *
   * @param s the text to append 
   */
  public void append(String s) {
    append(s,0);
  }

  /**
   * Append the passed <code>MathExpression</code> to the content.
   *
   * @param expr the <code>MathExpression</code> to append
   * @param actions action to while appending the text. One or more of 
   * {@link ScrollableTextAreaConstants#SET_MARK}, 
   * {@link ScrollableTextAreaConstants#NO_UPDATE},
   * {@link ScrollableTextAreaConstants#SCROLL}.
   * These can be strung together with bitwise OR, "|".
   */
  public synchronized void append(MathExpression expr, int actions) {
    //System.out.println("append m");

    if((actions & SET_MARK) == SET_MARK) {
      setScrollMark();
    }

    add(expr);

    if((actions & SCROLL) == SCROLL)
      scrollToMark();

    if((actions & NO_UPDATE) == 0)
      update();
  }

  /**
   * Append the passed <code>MathExpression</code> to the content.
   *
   * @param expr the <code>MathExpression</code> to append
   */
  public void append(MathExpression expr) {
    append(expr,0);
  }

  /**
   * Ensures that the scroll bars and text area are properly updated
   * for appearance.
   */
  protected void update() {
    //System.out.println("update");
    
    invalidate();
    validate();
    sp.invalidate();
    sp.validate();
    componentResized(null);
    //sp.repaint();
    sp.repaint();
  }

  /**
   * Returns the JScrollPane of this ScrollableTextArea
   *
   * @return the JScrollPane of this ScrollableTextArea
   */
  public JScrollPane getJScrollPane() {
    return sp;
  }  


  /**
   * Set the passed JTextArea's settings (font, line wrap, etc.) to
   * the correct settings for this ScrollableTextArea.
   *
   * @param ta a <code>JTextArea</code> value
   */
  protected void setTextAreaSettings(JTextArea ta) {
    ta.setFont(getFont());
    ta.setLineWrap(true);
    ta.setWrapStyleWord(true);
    ta.setEditable(false);
  }

   /**
   * Set the margins of this text box's JScrollPane to the given pixel
   * value
   *
   * @param m the pixel width of the margin
   */
  public void setMargins(int m) {
    sp.setViewportBorder(new LineBorder(getBackground(),m));
  } 
  

  /**
   * Set the scroll mark to the bottom of the current contents.
   *
   * @see #scrollToMark()
   */
  public void setScrollMark() {
    try {
      SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            scrollMark = getContentsHeight();
            //System.out.println("sm = " + scrollMark + " (cs = " + contents.size() +")");
          }
        });
    } catch (Exception e) {
      //do nothing.
    }
  }

  /**
   * Returns the height of the contents.
   *
   * @return the height of the contents.
   */
  public int getContentsHeight() {
    int m = 0;
    for(int i = 0; i < contents.size(); i++) {
      m += ((JComponent)contents.get(i)).getHeight();
    }

    //System.out.println("size = " + m + " contain = " + getSize().getHeight() + " num = " + contents.size());
    
    return m;
  }

    /**
   * Scroll the contents to the current scroll mark.
   *
   * @see #setScrollMark()
   */
  public void scrollToMark() {
    //we don't want any of this stuff to happen until everything is
    //validated. So, we put the scroll bar update on the event queue.
    update();
    update();


    SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          BoundedRangeModel sm = sp.getVerticalScrollBar().getModel();
          
          //System.out.println(sm.getMaximum() + ", " + sm.getMinimum() + ", " + sm.getExtent());
          sm.setValue
            (scrollMark);
          //System.out.println("scroll to: " + sm.getValue() + " (" + scrollMark + ")");
          
          sp.invalidate();
          sp.validate();
          repaint();
          //System.out.println("scrolled to " + scrollMark + 
          //                   " (" + getContentsHeight() + ")");
        }
      });
    repaint();
  }


  public void adjustmentValueChanged(AdjustmentEvent e) {
    /*
    JScrollBar sb = (JScrollBar) e.getSource();

    BoundedRangeModel sm = sb.getModel();

    if(!sb.getValueIsAdjusting()) {
      System.out.print("ae: " + sb.getValue() + " (");
      System.out.println(sm.getMaximum() + ", " + sm.getMinimum() + ", " + sm.getExtent() + ")");
      }
    */
  }
}