package vcp.utilities;

import java.awt.GraphicsEnvironment;
import java.awt.Font;
import java.awt.font.TextAttribute;
import java.util.Arrays;
import java.util.HashMap;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;

/**
 * A utility class with static methods to deal with Unicode.
 *
 * @author phil kaasa
 * @version 1.0
 */
public class UnicodeUtil {
  /** indicates a serif type font */
  public static final int SERIF = 1;

  /** indicates a sans-serif type font */
  public static final int SANS = 2;

  /** */
  public static final int PLAIN = Font.PLAIN;

  /** */
  public static final int ITALIC = Font.ITALIC;

  /** */
  public static final int BOLD = Font.BOLD;

  /** codepoints for mathematical operators*/
  private static final String mathText =
    generateUnicodeString(0x2200,0x228e);    //mathematical operators

  /** codepoints for upper and lower case Greek */
  private static final String greekText = 
    generateUnicodeString(0x03b1,0x03c9) +   //greek lower case
    generateUnicodeString(0x0391,0x03a1) +   //greek upper case 1
    generateUnicodeString(0x03a3,0x03a9);    //greek upper case 2

  /** codepoints for numbers plus super- and subscripts*/
  private static final String numberText =
    "1234567890" +
    "\u2070\u00b9\u00b2\u00b3" +
    generateUnicodeString(0x2074,0x2079) +  //superscripts
    generateUnicodeString(0x2080,0x2089);   //subscripts

  /** codepoints for alphanumerics, numberText, and greekText*/
  private static final String alphaNumberText = 
    generateUnicodeString(0x0020,0x007e) +
    generateUnicodeString(0x00a1,0x00ff) +
    numberText + greekText;

  /** codepoints for supplemental mathematical operators */
  private static final String supplementalText = 
    generateUnicodeString(0x228f,0x22ff);    //mathematical operators
    
  /** used for unicode data */
  private static HashMap unicodeData;

  /** font to use for */
  private static Font alphaNumberFont = null;
  private static Font greekFont = null;
  private static Font mathFont = null;
  
  /**
   * Returns a font from the given array which can display all the
   * characters in the given String.
   *
   * @param fonts the <code>Font</code>s to try
   * @param string the <code>String</code> of test characters
   * @return a <code>Font</code> that can display all the characters
   * in string
   * @exception NoFontFoundException if none of the given fonts can
   * display the test string.
   * @see #getDisplayFont(String, String[], String[])
   * @see #getDisplayFont(String, int)
   */
  public static Font getDisplayFont(String string, Font[] fonts) 
    throws NoFontFoundException {
    
    char[][] undisplayables = new char[fonts.length][];
    
    for(int i = 0; i < fonts.length; i++)
      if(fonts[i].canDisplayUpTo(string) == -1)
        return fonts[i];
      else {
        undisplayables[i] = getUndisplayableChars(string,fonts[i]);
      }
    
    if(fonts.length == 0)
      throw new NoFontFoundException
        ("None of the passed fonts are available on the system",
         fonts, undisplayables);
    else
      throw new NoFontFoundException
        ("None of the available fonts can display the entire " +
         "requested set of characters", 
         fonts, undisplayables);    
  }

  /**
   * Returns a font from the given arrays which can display all the
   * characters in the given String. Of the font families given, this
   * method will only test font families available on the system.
   *
   * @param string the <code>String</code> of test characters
   * @param toTryFamily an array of font family names (the family
   * names must correspond to the font name at the same index in
   * toTryName)
   * @param toTryName an array of font names (the font names must
   * correspond to the font family at the same index in toTryFamily)
   * @return a <code>Font</code> that can display all the characters
   * in string
   * @exception NoFontFoundException if none of the given fonts can
   * display the test string or if none of the fonts are available on
   * the system.
   * @see #getDisplayFont(String, Font[])
   * @see #getDisplayFont(String, int)
   */
  public static Font getDisplayFont(String string, String[] toTryFamily,
                                    String[] toTryName)
    throws NoFontFoundException {

    GraphicsEnvironment ge = 
      GraphicsEnvironment.getLocalGraphicsEnvironment();
    
    String [] sysFonts = ge.getAvailableFontFamilyNames();
    Arrays.sort(sysFonts);

    Font[] fontsT = new Font[toTryName.length];
    int index = 0;
    
    for(int i = 0; i < toTryFamily.length; i++)
      if(Arrays.binarySearch(sysFonts, toTryFamily[i]) >= 0)
        fontsT[index++] = new Font(toTryName[i],Font.PLAIN,1);
    
    Font[] fonts = new Font[index];

    System.arraycopy(fontsT,0,fonts,0,index);
    
    return getDisplayFont(string, fonts);
  }

  /**
   * Returns a <code>Font</code> of the given type which can display
   * the test string.
   *
   * @param type SERIF or SANS
   * @return a <code>Font</code> that can display all the characters
   * in string
   * @exception NoFontFoundException if no font of the given type can
   * be found that can display the test string.
   * @exception IllegalArgumentException if <code>type</code> is not
   * SERIF or SANS
   * @see #getDisplayFont(String, Font[])
   * @see #getDisplayFont(String, String[], String[])
   */
  public static Font getDisplayFont(String string, int type)
    throws NoFontFoundException {
    if(type == SERIF)
      return getDisplayFont(string,
                            new String[] {"Serif",
                                          "Lucidia Bright",
                                          "Times New Roman", 
                                          "Hiragino Mincho Pro",
                                          "Apple Symbols"},
                            
                            new String[] {"Serif",
                                          "Lucdia Bright Regular",
                                          "Times New Roman", 
                                          "HiraMinPro-W3",
                                          "Apple Symbols"});    

    else if (type == SANS)
      return getDisplayFont(string,
                            new String[] {"Sans",
                                          "Lucida Sans",
                                          "Arial",
                                          "Hiragino Maru Gothic Pro",
                                          "Apple Symbols"},
                            
                            new String[] {"Sans",
                                          "Lucida Sans",
                                          "Arial",
                                          "HiraMaruPro-W4",
                                          "Apple Symbols"});
    
    else
      throw new IllegalArgumentException("Type is invalid: " + type);
    
  }


  /**
   * Returns a serif style <code>Font</code> which can display the
   * Unicode code points U+2200 - U+228E.
   *
   * @return a <code>Font</code> which can display the Unicode code
   * points U+2200 - U+228E
   * @exception NoFontFoundException if no <code>Font</code> can be
   * found to display the code points.
   */
  public static Font getMathematicalFont() throws NoFontFoundException {
    return getMathematicalFont(SERIF);
  }

  /**
   * Returns a <code>Font</code> of the given type which can display
   * the Unicode code points U+2200 - U+228E.
   *
   * @param type SERIF or SANS
   * @return a <code>Font</code> value
   * @exception NoFontFoundException if no <code>Font</code> can be
   * found to display the code points.
   * @exception IllegalArgumentException if <code>type</code> is not
   * SERIF or SANS
   */
  public static Font getMathematicalFont(int type) throws NoFontFoundException {
    if(mathFont == null)
      return mathFont = getDisplayFont(mathText, type);
    else
      return mathFont;
  }

  /**
   * Returns a serif style <code>Font</code> which can display the
   * Unicode code points U+0391 - U+03A1, U+03A3 - U+03A9, and U+03B1
   * - U+03C9 (Greek upper and lower case letters).
   *
   * @return a <code>Font</code> which can display Greek upper and
   * lower case letters.
   * @exception NoFontFoundException if no <code>Font</code> can be
   * found to display the code points.
   */
  public static Font getGreekFont() throws NoFontFoundException {
    return getGreekFont(SERIF);
  }

  /**
   * Returns a <code>Font</code> of the given type which can display
   * the Unicode code points U+0391 - U+03A1, U+03A3 - U+03A9, and
   * U+03B1 - U+03C9 (Greek upper and lower case letters).
   *
   * @param type SERIF or SANS
   * @return a <code>Font</code> which can display Greek upper and
   * lower case letters.
   * @exception NoFontFoundException if no <code>Font</code> can be
   * found to display the code points.
   * @exception IllegalArgumentException if <code>type</code> is not
   * SERIF or SANS
   */
  public static Font getGreekFont(int type) throws NoFontFoundException {
    if(greekFont ==null)
      return greekFont = getDisplayFont(greekText, type);
    else
      return greekFont;
  }

  /**
   * Returns a serif type <code>Font</code> which can display the
   * Unicode Basic Latin and Latin-1 Supplement blocks, the Unicode
   * code points U+0391 - U+03A1, U+03A3 - U+03A9, U+03B1 - U+03C9
   * (Greek upper and lower case letters), and numeric sub- and
   * superscripts.
   *
   * @return a serif type <code>Font</code>
   * @exception NoFontFoundException if no <code>Font</code> can be
   * found to display the code points.
   */
  public static Font getAlphaNumberFont() throws NoFontFoundException {
    return getAlphaNumberFont(SERIF);
  }

  /**
   * Returns a <code>Font</code> of the given type which can display
   * the Unicode Basic Latin and Latin-1 Supplement blocks, the
   * Unicode code points U+0391 - U+03A1, U+03A3 - U+03A9, U+03B1 -
   * U+03C9 (Greek upper and lower case letters), and numeric sub- and
   * superscripts.
   *
   * @param type SERIF or SANS
   * @return a <code>Font</code> of the given type
   * @exception NoFontFoundException if no <code>Font</code> can be
   * found to display the code points.
   * @exception IllegalArgumentException if <code>type</code> is not
   * SERIF or SANS
   */
  public static Font getAlphaNumberFont(int type) throws NoFontFoundException {
    if(alphaNumberFont == null)
      return alphaNumberFont = getDisplayFont(alphaNumberText, type);
    else
      return alphaNumberFont;
  }

  /**
   * Returns an array of <code>char</code>s which contains characters
   * in <code>s</code> which cannot be displayed by the given
   * <code>Font</code>.
   *
   * @param s the test String
   * @param font the <code>Font</code> to test
   * @return an array of <code>char</code>s
   */
  public static char[] getUndisplayableChars(String s, Font font) {
    StringBuffer sb = new StringBuffer();
    char[] test = s.toCharArray();
    int i = 0;

    while((i = font.canDisplayUpTo(test,i,test.length)) >= 0) {
      sb.append(test[i++]);
    }

    return (sb.toString().toCharArray());
  }
  
  /**
   * Returns a String containing the given range of Unicode code
   * points.
   *
   * @param li lower bound (inclusive)
   * @param hi upper bound (inclusive)
   * @return a String containing the given range of Unicode code
   * points.
   */
  private static String generateUnicodeString(int li, int hi) {
    char l = (char) li;
    char h = (char) hi;

    StringBuffer sb = new StringBuffer();

    for(char i = l; i <= h; i++) {
      sb.append(i);
    }

    return sb.toString();
  }

  /**
   * Initialize the Unicode data for the code points corresponding to
   * the given <code>String</code>
   *
   * @param codePoints the <code>String</code> of code points to
   * initialize data for
   */
  public static void initializeUnicodeData(String codePoints) {
    initializeUnicodeData(codePoints.toCharArray());
  }

  /**
   * Initialize the Unicode data for the given range of code points.
   *
   * @param low lower bound (inclusive)
   * @param high upper bound (inclusive)
   */
  public static void initializeUnicodeData(int low, int high) {
    int[] cp = new int[high - low + 1];

    for(int val = low, i = 0; val <= high; val++, i++) {
      cp[i] = val;
    }

    initializeUnicodeData(cp);
  }

  /**
   * Initialize the Unicode data for the code points corresponding to
   * the given <code>char</code>s.
   *
   * @param chars the <code>char</code>s to initialize data for
   */
  public static void initializeUnicodeData(char[] chars) {
    int[] cp = new int[chars.length];

    for(int i = 0; i < chars.length; i++)
      cp[i] = (int) chars[i];

    initializeUnicodeData(cp);
  }

  /**
   * Initialize the Unicode data for the given code points.
   *
   * @param codePoints the code points to initialize data for
   */
  public static void initializeUnicodeData(int[] codePoints) {
    BufferedReader br;
    
    unicodeData = new HashMap(codePoints.length);

    Arrays.sort(codePoints);
    
    String[] cp;
    try {
      br = new BufferedReader
        (new InputStreamReader
         (UnicodeUtil.class.getResourceAsStream
          ("/vcp/utilities/UnicodeData.txt")));
      for(int i = 0; i < codePoints.length; i++) {
        while(Integer.parseInt
              ((cp = br.readLine().split(";"))[0],16) 
              != codePoints[i]);
        unicodeData.put(new Integer(codePoints[i]),
                        cp[1]);
      }

    } catch (IOException ioe) {
      ExceptionDialog.showExceptionDialog
        ("Unicode Data Error",
         "There was an error loading the Unicode data.",
         ioe);
    }
  }

  /**
   * Returns the description of the passed code point.
   *
   * @param codePoint a Unicode code point
   * @return the description of the passed code point
   */
  public static String getUnicodeData(int codePoint) {
    return (String) unicodeData.get(new Integer(codePoint));
  }

  /**
   * Returns a String which is the passed String with all characters
   * that cannot be displayed by the passed font replaced with the
   * corresponding Unicode description.
   *
   * @param string the string to fix
   * @param font the font to display the string
   * @return a String with undisplayable characters replaced
   */
  public static String fixString(String string, Font font) {
    char[] bc = getUndisplayableChars(string,font);
    
    if(bc.length > 0) {
      String copy = new String(string);
      
      initializeUnicodeData(bc);
      
      //System.out.println(unicodeData);
      
      for(int i = 0; i < bc.length; i++) {
        copy = copy.replaceAll
          (bc[i]+"",
           "[" + getUnicodeData((int) bc[i]) + "]");
      }
      
      return copy;
    }
    
    else
      return string;
  }

  public static void main(String[] args) {
    /*
    try {
      getDisplayFont("abcd\u2084\uffff\u2204", SERIF);
      System.out.println("yes");
    } catch (NoFontFoundException e) {
      for(int j = 0; j < e.getFontList().length; j++) {
        System.out.println(e.getFontList()[j] + ": " + e.getBadChars(j).length);

        for(int i = 0; i < t.length; i++)
          System.out.println(Integer.toString((int)t[i],16));

      }

      System.out.println(e.getFontList()[e.getBestFont()]);
      

    }
    */

    //initializeUnicodeData(new int[] {0x03bb,0x0020,0x0030,0x2204});
    //printUnicodeData();

    System.out.println(fixString("This is text with broken characters like \u03bb. It can't display nice stuff: {v\u2081, v\u2082}.", new Font("Serif",Font.PLAIN,1)));
    
    /*
    System.out.print("Mathematical: ");
    try {
      font = getMathematicalFont();
    } catch (Exception e) {
      System.out.println("no");
      System.out.println(e.getMessage());
    } finally {
      System.out.println(font.getName());
    }

    System.out.print("Greek: ");
    try {
      font = getGreekFont();
    } catch (Exception e) {
      System.out.println("no");
      System.out.println(e.getMessage());
    } finally {
      System.out.println(font.getName());
    }

    System.out.print("AlphaNumber: ");
    try {
      font = getAlphaNumberFont();
    } catch (Exception e) {
      System.out.println("no");
      System.out.println(e.getMessage());
    } finally {
      System.out.println(font.getName());
    }
    */

  }
}